"""
A Pythonic way to analyze and work with Beam Delivery SIMulation (BDSIM) ROOT output files.

Design goals:
 - No dependency on (py)ROOT(py) is needed. The module uses `uproot` instead.
 - Enables and favors exploration of the ROOT files. No prior knowledge of the content should be required
 to explore and discover the data structure.
 - provide analysis tools exploiting the new Awkward 1.0 library (https://arxiv.org/pdf/2001.06307.pdf)
"""
# this must apparently be at the beginning of the file or the imports crash
from __future__ import annotations as _annotations

from . import Data as _Data

import awkward as _ak
from collections import UserDict as _UserDict
import logging as _logging
import itertools as _itertools
import os as _os
import numpy as _np
import pandas as _pd
from scipy.interpolate import interp1d as _interp1d
from typing import Optional as _Optional
from typing import List as _List
from typing import Dict as _Dict
from typing import Tuple as _Tuple
from typing import Union as _Union
import matplotlib.pyplot as _plt
import warnings as _warnings

try:
    import uproot as _uproot
except (ImportError, ImportWarning):
    _logging.warning("Uproot is required for this module to have full functionalities.\n")
    raise ImportError("Uproot is required for this module to work.")

_WITH_ROOT = False
try:
    import ROOT as _ROOT
    _WITH_ROOT = True
except ImportError:
    pass
    #_logging.warning("ROOT is required for this module to have full functionalities.\n"
    #                 "Not all methods will be available.")

_WITH_BOOST_HISTOGRAM = False
try:
    try:
        _warnings.simplefilter("ignore")
        import boost_histogram as bh
        _warnings.simplefilter("default")
    except (ImportError, UserWarning):
        pass
    _WITH_BOOST_HISTOGRAM = True
except (ImportError, ImportWarning):
    _logging.warning("boost_histogram is required for this module to have full functionalities.\n"
                     "Not all methods will be available.")

__all__ = [
    'Output',
    'BDSIMOutputException',
    'BDSIMOutput',
    'REBDSIMOutput',
    'REBDSIMOpticsOutput',
    'REBDSIMCombineOutput',
    'Histogram',
    'Histogram2D',
    'Histogram3D'
]


class BDSIMOutputException(Exception):
    pass


class Histogram:
    def __init__(self, h):  # h is a THxD
        self._h = h
        self.variances = h.variances()
        self._centers = None
        self._normalized_values = None
        self.normalization = 1.0
        self.coordinates_normalization = 1.0

    def __getattr__(self, item):
        return getattr(self._h, item)

    def set_normalization(self, normalization: float = 1.0):
        self.normalization = normalization
        self._normalized_values = self.normalization * self._h.values
        return self

    def set_coordinates_normalization(self, normalization: float = 1.0):
        self.coordinates_normalization = normalization
        return self

    @property
    def normalized_values(self):
        return self._normalized_values

    @property
    def xnumbins(self):
        return len(self._h.axes[0].edges()) - 1


class Histogram1D(Histogram):
    @property
    def centers(self):
        if self._centers is not None:
            return self._centers
        self._centers = self.axes[0].centers()
        return self._centers

    @property
    def values(self):
        return self._h.values()


class Histogram2D(Histogram):
    ...


class Histogram3D(Histogram):
    def __init__(self, h, parent, name):
        Histogram.__init__(self, h)
        self._filename = parent.filename
        self._path = parent.path
        self._name = name
        self._meshname = self._name.split('-')[0]
        self._parent = parent

    @property
    def filename(self):
        return self._filename

    @property
    def path(self):
        return self._path

    @property
    def meshname(self):
        return self._meshname

    @property
    def values(self):
        return self._h.values()

    @property
    def bins_volume(self):
        return _np.diff(self._h.axes[0].edges())[0] * \
               _np.diff(self._h.axes[1].edges())[0] * \
               _np.diff(self._h.axes[2].edges())[0]

    @property
    def edges(self):
        return _np.array([list(self._h.axes[0].edges()),
                          list(self._h.axes[1].edges()),
                          list(self._h.axes[2].edges())])

    @property
    def bins(self):
        bins_x = _np.diff(self.bh.axes[0].edges)[0]
        bins_y = _np.diff(self.bh.axes[1].edges)[0]
        bins_z = _np.diff(self.bh.axes[2].edges)[0]
        return _np.array([bins_x, bins_y, bins_z], dtype=object)

    @property
    def centers(self):
        if self._centers is not None:
            return self._centers
        self._centers = [[
            self.coordinates_normalization * (self.edges[j][i] + self.edges[j][i + 1]) / 2
            for i in range(len(self.edges[j]) - 1)] for j in range(3)]
        return self._centers

    @property
    def ynumbins(self):
        return len(self._h.axes[1].edges()) - 1

    @property
    def znumbins(self):
        return len(self._h.axes[2].edges()) - 1

    @property
    def scoring_mesh_translations(self):
        dico = self._parent.model.scoring_mesh_translations[self._meshname]
        return [dico['fX'], dico['fY'], dico['fZ']]

    @property
    def scoring_mesh_rotations(self):
        return self._parent.model.scoring_mesh_rotations[self._meshname]

    def to_df(self):
        index = _pd.MultiIndex.from_product(self.centers, names=('X', 'Y', 'Z'))
        data = {
            'edep': self.normalized_values.flatten() / self.bins_volume,
        }
        return _pd.DataFrame(index=index, data=data)

    def to_csv(self, filename='histogram.csv', path='.', **kwargs):
        self.to_df().to_csv(_os.path.join(path, filename), header=False, float_format='% 11.7E', **kwargs)


class Histogram4D:
    def __init__(self, parent, bdsbh4d_histogram, name):
        self._h = bdsbh4d_histogram
        self._bh = None
        self._bh_error = None
        self._energy_axis_type = name.split('-')[-1]
        self._filename = parent.filename
        self._path = parent.path
        self._meshname = name.split('-')[0]
        self._cache = None
        self._coordinates_normalization = 1.0
        self._parent = parent

    @property
    def scoring_mesh_translations(self):
        dico = self._parent.model.scoring_mesh_translations[self._meshname]
        return [dico['fX'], dico['fY'], dico['fZ']]

    @property
    def scoring_mesh_rotations(self):
        return self._parent.model.scoring_mesh_rotations[self._meshname]

    def get_pyboost(self, hist_type):
        if _WITH_BOOST_HISTOGRAM:
            energy_axis = None
            if self._energy_axis_type == 'log':
                energy_axis = bh.axis.Regular(self._h.h_nebins, self._h.h_emin, self._h.h_emax,
                                              transform=bh.axis.transform.log)
            elif self._energy_axis_type == 'linear':
                energy_axis = bh.axis.Regular(self._h.h_nebins, self._h.h_emin, self._h.h_emax)
            elif self._energy_axis_type == 'user':
                energy_axis = bh.axis.Variable(self._h.h_ebinsedges)

            histo4d = bh.Histogram(
                bh.axis.Regular(self._h.h_nxbins, self._h.h_xmin, self._h.h_xmax),
                bh.axis.Regular(self._h.h_nybins, self._h.h_ymin, self._h.h_ymax),
                bh.axis.Regular(self._h.h_nzbins, self._h.h_zmin, self._h.h_zmax),
                energy_axis)

            for x, y, z, e in _itertools.product(range(self._h.h_nxbins), range(self._h.h_nybins),
                                                 range(self._h.h_nzbins), range(self._h.h_nebins)):
                histo4d[x, y, z, e] = getattr(self._h, hist_type).at(x, y, z, e)

            return histo4d
        else:
            raise AttributeError("Boost histograms are not available")

    def extract_spectrum(self, x=0, y=0, z=0, path='.', extract_all=False):

        if not extract_all:
            all_x = [x]
            all_y = [y]
            all_z = [z]
        else:
            all_x = range(self.xnumbins)
            all_y = range(self.ynumbins)
            all_z = range(self.znumbins)

        for _x in all_x:
            for _y in all_y:
                for _z in all_z:

                    f = open(f"{path}/fluxes_{self.meshname}_{_x}_{_y}_{_z}", 'w')
                    spectrum = list(self.bh[_x, _y, _z, :].to_numpy()[0])
                    spectrum.reverse()

                    i = 1
                    for value in spectrum:
                        f.write("  {:.4E}".format(value))
                        if i % 6 == 0 and i != len(spectrum):
                            f.write('\n')
                        i += 1
                    f.write('\n 1.000\n')
                    f.write(f'fluxes_{self.meshname}_{_x}_{_y}_{_z}')

                    f.close()

    def extract_error(self, x=0, y=0, z=0, path='.', extract_all=False):

        if not extract_all:
            all_x = [x]
            all_y = [y]
            all_z = [z]
        else:
            all_x = range(self.xnumbins)
            all_y = range(self.ynumbins)
            all_z = range(self.znumbins)

        for _x in all_x:
            for _y in all_y:
                for _z in all_z:

                    f = open(f"{path}/errors_{self.meshname}_{_x}_{_y}_{_z}", 'w')
                    spectrum = list(self.bh_err[_x, _y, _z, :].to_numpy()[0])
                    spectrum.reverse()

                    i = 1
                    for value in spectrum:
                        f.write("  {:.4E}".format(value))
                        if i % 6 == 0 and i != len(spectrum):
                            f.write('\n')
                        i += 1
                    f.write('\n 1.000\n')
                    f.write(f'errors_{self.meshname}_{_x}_{_y}_{_z}')

                    f.close()

    def project_to_3d(self, weights=1):

        histo3d = bh.Histogram(*self.bh.axes[0:3])

        for x in range(self.bh.shape[0]):
            for y in range(self.bh.shape[1]):
                for z in range(self.bh.shape[2]):
                    tmp = self.bh[x, y, z, :].view() * weights
                    histo3d[x, y, z] = tmp.sum()

        self._cache = histo3d.view()
        return histo3d

    def compute_h10(self, conversionfactorfile):
        data = _pd.read_table(conversionfactorfile, names=["energy", "h10_coeff"])
        f = _interp1d(data['energy'].values, data['h10_coeff'].values)
        self._cache = self.project_to_3d(weights=f(self.bh.axes[3].centers)).view()
        return self.project_to_3d(weights=f(self.bh.axes[3].centers))

    def plot_spectrum(self, x=0, y=0, z=0, xlim=[1e-10, 300], ylim=[1e-15, 1e-7], log=True, error_bar=True, **kwargs):

        fig, ax = _plt.subplots(sharex=True, figsize=(8, 5.5))

        bh = self.bh
        bh_err = self.bh_err

        bins = bh.axes[3].edges
        centers = bh.axes[3].centers
        widths = bh.axes[3].widths

        values = bh[x, y, z, :].to_numpy()[0]

        energies = []
        spectrum = []

        scale = 1000

        for j in range(len(bins)):

            if j == 0:
                energies.append(bins[j] * scale)
                spectrum.append(values[j])
            if j == len(values):
                energies.append(bins[j] * scale)
                spectrum.append(values[j - 1])
            else:
                energies.append(bins[j] * scale)
                spectrum.append(values[j - 1])

                energies.append(bins[j] * scale)
                spectrum.append(values[j])

        ax.plot(energies, spectrum, **kwargs)

        if error_bar:
            ax.bar(centers * scale, bh[x, y, z, :].to_numpy()[0], widths, yerr=bh_err[x, y, z, :].to_numpy()[0],
                   color='white', edgecolor='white', error_kw=dict(ecolor='black', lw=1))

        ax.set_xlim(xlim[0], xlim[1])

        ax.set_xlabel("Energy (MeV)", fontsize=15, fontweight='bold')
        ax.set_ylabel("Flux (cm$^{-2}$ primary$^{-1}$)", fontsize=15, fontweight='bold')

        if log:
            ax.set_xscale('log')
            ax.set_yscale('log')
            ax.set_ylim(ylim[0], ylim[1])
        else:
            ax.set_ylim(ylim[0], ylim[1])

        ax.legend(loc=0, fontsize=12.2)
        ax.grid(b=None)

        fig.tight_layout()

        _plt.show()

    def get_position_in_space(self, x=0, y=0, z=0, origin=[0, 0, 0]):
        centers = self.centers
        pos_x = origin[0] + centers[0][x]
        pos_y = origin[1] + centers[1][y]
        pos_z = origin[2] + centers[2][z]
        return _np.array([pos_x, pos_y, pos_z], dtype=object)

    def set_values(self, new_values):
        self._cache = new_values

    @property
    def filename(self):
        return self._filename

    @property
    def path(self):
        return self._path

    @property
    def meshname(self):
        return self._meshname

    @property
    def bh(self):
        if self._bh is None:
            self._bh = self.get_pyboost('h')
        return self._bh

    @property
    def bh_err(self):
        if self._bh_error is None:
            self._bh_error = self.get_pyboost('h_err')
        return self._bh_error

    @property
    def h(self):
        return self._h.h

    @property
    def h_err(self):
        return self._h.h_err

    @property
    def values(self):
        return self._cache

    @property
    def xnumbins(self):
        return self.bh.axes[0].size

    @property
    def ynumbins(self):
        return self.bh.axes[1].size

    @property
    def znumbins(self):
        return self.bh.axes[2].size

    @property
    def enumbins(self):
        return self.bh.axes[3].size

    @property
    def shape(self):
        return (self.xnumbins, self.ynumbins, self.znumbins, self.enumbins)

    @property
    def find_max_coordinates(self):
        result = _np.where(self.values == _np.max(self.values))
        return result[0][0], result[1][0], result[2][0]

    @property
    def coordinates_normalization(self):
        return self._coordinates_normalization

    @property
    def bins(self):
        bins_x = _np.diff(self.bh.axes[0].edges)[0]
        bins_y = _np.diff(self.bh.axes[1].edges)[0]
        bins_z = _np.diff(self.bh.axes[2].edges)[0]
        return _np.array([bins_x, bins_y, bins_z], dtype=object)

    @property
    def bins_volume(self):
        return _np.diff(self.bh.axes[0].edges)[0] * \
               _np.diff(self.bh.axes[1].edges)[0] * \
               _np.diff(self.bh.axes[2].edges)[0]

    @property
    def edges(self):
        edges_x = self.bh.axes[0].edges
        edges_y = self.bh.axes[1].edges
        edges_z = self.bh.axes[2].edges
        edges_e = self.bh.axes[3].edges
        return _np.array([edges_x, edges_y, edges_z, edges_e], dtype=object)

    @property
    def centers(self):
        centers_x = self.bh.axes[0].centers
        centers_y = self.bh.axes[1].centers
        centers_z = self.bh.axes[2].centers
        centers_e = self.bh.axes[3].centers
        return _np.array([centers_x, centers_y, centers_z, centers_e], dtype=object)


class OutputType(type):
    """A generic type for BDSIM output classes."""
    pass


class Output(metaclass=OutputType):
    def __init__(self, filename: str = 'output.root', path: str = '.', *, open_file: bool = True):
        """
        Create a representation of a BDSIM output using uproot to read the root file.

        The root file is opened with uproot, so a valid path and filename must be provided.

        Args:
            filename: the name of the root file to read
            path: the path to the root file
            open_file: attempts to open the master file
        """
        self._filename = filename
        self._path = path
        self._file = _os.path.join(path, filename)
        if open_file:
            self._root_directory: _uproot.rootio.ROOTDirectory = _uproot.open(self._file)
            if _WITH_ROOT:
                _Data.LoadROOTLibraries()
                self._rootfile = _ROOT.TFile.Open(self._file)


    @classmethod
    def from_root_directory(cls, directory: _uproot.rootio.ROOTDirectory) -> Output:
        """Create an `Output` object directly attached to an existing ROOT directory.

        Args:
            directory: an existing `uproot` `ROOTDirectory`
        """
        o = cls(open_file=False)
        o._file = None
        o._root_directory = directory
        return o

    def __getitem__(self, item: str):
        """Read an object from the ROOT file or directory by name."""
        return self._root_directory[item]

    @property
    def directory(self) -> _uproot.rootio.ROOTDirectory:
        """Return the master directory attached to this parent."""
        return self._root_directory

    @property
    def file(self):
        return self._file

    @property
    def filename(self):
        return self._filename
    @property
    def rootfile(self):
        return self._rootfile

    @property
    def path(self):
        return self._path


    class Directory:
        def __init__(self, parent: _Union[Output, Output.Directory], directory: _uproot.rootio.ROOTDirectory):
            """
            A representation of a (nested) structure of ROOT directories.

            Args:
                parent: the `Output` to which the directory structure is attached
                directory: the top-level ROOT directory
            """

            self._output: Output = parent
            self._directory: _uproot.rootio.ROOTDirectory = directory
            for name, cls in self._directory.iterclassnames(recursive=False):
                setattr(self, name.split(';')[0].replace('-', '_'), self.build(name, cls))

        def __getitem__(self, item):
            return self._directory[item]

        def build(self, n, c):
            if c.endswith('Directory'):
                item = self._directory[n]
                return Output.Directory(self.parent, directory=item)
            elif c.endswith('TH1D'):
                item = self._directory[n]
                return Histogram1D(item)
            elif c.endswith('TH2D'):
                item = self._directory[n]
                return Histogram2D(item)
            elif c.endswith('TH3D'):
                item = self._directory[n]
                return Histogram3D(item, self.parent, n)
            elif c.startswith('BDSBH4D'):
                name = n.split(';')[0]
                if not _WITH_ROOT:
                    _logging.error("ROOT must be installed to use Histogram4D.")
                else:
                    return Histogram4D(self.parent, self.parent.rootfile.Get('Event/MergedHistograms/' + name), name)
            else:
                return self._directory[n]

        def get(self, name):
            for n, c in self._directory.iterclassnames(recursive=False):
                if n == name:
                    return self.build(n, c)

        @property
        def parent(self) -> Output:
            """The parent Output to which the directory structure is attached."""
            return self._output

        @property
        def root_directory(self) -> _uproot.rootio.ROOTDirectory:
            """The associated uproot directory."""
            return self._directory

        @property
        def keys(self) -> _List[str]:
            """The content of the directory."""
            return self._directory.keys()

    class Tree:
        def __init__(self, parent: Output, tree_name: str = None):
            """
            A representation of a ROOT TTree structure.

            Args:
                parent: the `Output` to which the parent is attached
            """
            self._parent = parent
            self._tree_name = tree_name or self.__class__.__name__
            self._tree: _uproot.rootio.TTree = parent[self._tree_name]
            self._df: _Optional[_pd.DataFrame] = None
            self._np: _Optional[_np.ndarray] = None

        def __getitem__(self, item):
            try:
                return self._tree[item]
            except KeyError:
                return self._tree[item + '.']

        def __getattr__(self, b):
            branch_class = getattr(self.__class__, b.title().replace('_', ''), None)
            if branch_class is not None:
                _ = branch_class(parent=self, branch_name=self.__class__.__name__)
                setattr(self, b, _)
                return getattr(self, b)
            else:
                raise AttributeError(f"Branch {b} does not exist for {self._tree_name}")

        def array(self, branch=None, **kwargs) -> _np.ndarray:
            """A proxy for the `uproot` `array` method."""
            return self._tree[self.__class__.__name__ + '.'][self.__class__.__name__ + '.' + branch].array().tolist()[0]

        def arrays(self, branches=None, **kwargs):
            """A proxy for the `uproot` `arrays` method."""
            if self._tree_name != 'Optics':
                return [self._tree[self.__class__.__name__ + '.'][self.__class__.__name__ + '.' + b].array().tolist()[0] for
                        b in branches]
            else:
                return [self._tree[b].array().tolist() for b in branches]

        def pandas(self, branches=None, **kwargs):
            """A proxy for the `uproot` `pandas` method."""
            return self._tree.pandas.df(branches=branches, **kwargs)

        def to_df(self) -> _pd.DataFrame:
            pass

        def to_np(self) -> _np.ndarray:
            pass

        @property
        def parent(self):
            """The parent Output to which the parent structure is attached."""
            return self._parent

        @property
        def tree(self) -> _uproot.rootio.TTree:
            """The associated uproot parent."""
            return self._tree

        @property
        def branches_names(self) -> _List[str]:
            return [b for b in self.tree.keys() if '/' not in b]

        @property
        def branches(self) -> _List[_uproot.rootio.TBranchElement]:
            return [b for b in self.tree.values() if b.name[-1] == '.']

        @property
        def numentries(self) -> int:
            """Provides the number of entries in the parent (without reading the entire file)."""
            return self.tree.num_entries

        @property
        def df(self) -> _pd.DataFrame:
            if self._df is None:
                return self.to_df()
            return self._df

        @property
        def np(self) -> _np.ndarray:
            if self._np is None:
                return self.to_np()
            return self._np

    class Branch:
        BRANCH_NAME: _Optional[str] = None
        DEFAULT_LEAVES: _Dict[str, _Tuple[bool, _Optional[str]]] = {}

        def __new__(cls, *args, **kwargs):
            def toggle(leave):
                def do_toggle(self):
                    self._active_leaves[leave][0] = not self._active_leaves[leave][0]
                    return self

                return do_toggle

            instance = super().__new__(cls)
            for k in cls.DEFAULT_LEAVES:
                setattr(instance, f"toggle_{k}", toggle(k).__get__(instance))
            return instance

        def __init__(self, parent: Output.Tree, branch_name: _Optional[str] = None):
            """
            A representation of a ROOT Branch.

            Args:
                parent: the `Tree` to which the branch is attached
            """
            self._parent: Output.Tree = parent
            if branch_name != "Event":
                self._branch = parent[branch_name]
            else:
                self._branch = parent[self.__class__.__name__]
            self._df: _Optional[_pd.DataFrame] = None
            self._np: _Optional[_np.ndarray] = None
            self._active_leaves: _Dict[str, _Tuple[bool, _Optional[str]]] = self.DEFAULT_LEAVES.copy()

        def __getitem__(self, item):
            return self._branch[item]

        def array(self, branch=None, **kwargs) -> _np.ndarray:
            """A proxy for the `uproot` `array` method."""
            return self.parent.array(branch=branch, **kwargs)

        def arrays(self, branches=None, **kwargs):
            """A proxy for the uproot method.
            """
            if branches is None:
                branches = self._active_leaves
            return self.parent.arrays(branches=[b for b in branches], **kwargs)

        def pandas(self, branches: _List[str] = None, strip_prefix: bool = True, **kwargs):
            """A proxy for the uproot method.
            """
            if branches is None:
                branches = [self.branch_name + b for b, _ in self._active_leaves.items() if _[0] is True]
            else:
                branches = [self.branch_name + b for b in branches]

            df = _ak.to_dataframe(self.parent.tree.arrays(branches, library='ak'),
                                  how='outer') # Use ak instead of pandas
            if strip_prefix:
                import re
                df.columns = [re.split(self.branch_name, c)[1] for c in df.columns]
            return df

        def to_df(self) -> _pd.DataFrame:
            if self._df is None:
                self._df = self.pandas()
            return self._df

        def to_np(self) -> _np.ndarray:
            pass

        @property
        def parent(self) -> Output.Tree:
            """The parent `Tree` to which the branch is attached."""
            return self._parent

        @property
        def branch(self) -> _uproot.rootio.TBranch:
            return self._branch

        @property
        def branch_name(self) -> str:
            if self.BRANCH_NAME == '':
                return ''
            name = self.branch.name
            if self.BRANCH_NAME:
                name += self.BRANCH_NAME
            if not name.endswith('.'):
                return name + '.'
            else:
                return name

        @property
        def leaves(self) -> _List[_uproot.rootio.TBranchElement]:
            return self._branch.values()

        @property
        def leaves_names(self) -> _List[str]:
            return self._branch.keys()

        @property
        def df(self) -> _pd.DataFrame:
            if self._df is None:
                return self.to_df()
            return self._df

        @property
        def np(self) -> _np.ndarray:
            if self._np is None:
                return self.to_np()
            return self._np


class BDSIMOutput(Output):
    def __getattr__(self, item):
        if item in (
                'header',
                'geant4data',
                'beam',
                'options',
                'model',
                'run',
                'event'
        ):
            setattr(self,
                    item,
                    getattr(BDSIMOutput, item.title())(parent=self)
                    )
            return getattr(self, item)

    class Header(Output.Tree):

        class Header(Output.Branch):
            DEFAULT_LEAVES = {
                'bdsimVersion': [True, None],
                'geant4Version': [True, None],
                'rootVersion': [True, None],
                'clhepVersion': [True, None],
                'timeStamp': [True, None],
                'fileType': [True, None],
                'dataVersion': [True, None],
                'doublePrecisionOutput': [True, None],
                'analysedFiles': [True, None],
                'combinedFiles': [True, None],
                'nTrajectoryFilters': [True, None],
                'trajectoryFilters': [True, None],
            }

    class Beam(Output.Tree):
        class BeamBase(Output.Branch):
            BRANCH_NAME = 'GMAD::BeamBase'
            DEFAULT_LEAVES = {
                'particle': [True, None],
                'beamParticleName': [True, None],
                'beamEnergy': [True, None],
                'beamKineticEnergy': [True, None],
                'beamMomentum': [True, None],
                'distrType': [True, None],
                'xDistrType': [True, None],
                'yDistrType': [True, None],
                'zDistrType': [True, None],
                'distrFile': [True, None],
                'distrFileFormat': [True, None],
                'matchDistrFileLength': [True, None],
                'nlinesIgnore': [True, None],
                'nlinesSkip': [True, None],
                'X0': [True, None],
                'Y0': [True, None],
                'Z0': [True, None],
                'S0': [True, None],
                'Xp0': [True, None],
                'Yp0': [True, None],
                'Zp0': [True, None],
                'T0': [True, None],
                'E0': [True, None],
                'Ek0': [True, None],
                'P0': [True, None],
                'tilt': [True, None],
                'sigmaT': [True, None],
                'sigmaE': [True, None],
                'sigmaEk': [True, None],
                'sigmaP': [True, None],
                'betx': [True, None],
                'bety': [True, None],
                'alfx': [True, None],
                'alfy': [True, None],
                'emitx': [True, None],
                'emity': [True, None],
                'dispx': [True, None],
                'dispy': [True, None],
                'dispxp': [True, None],
                'dispyp': [True, None],
                'emitNX': [True, None],
                'emitNY': [True, None],
                'sigmaX': [True, None],
                'sigmaXp': [True, None],
                'sigmaY': [True, None],
                'sigmaYp': [True, None],
                'envelopeX': [True, None],
                'envelopeXp': [True, None],
                'envelopeY': [True, None],
                'envelopeYp': [True, None],
                'envelopeT': [True, None],
                'envelopeE': [True, None],
                'envelopeR': [True, None],
                'envelopeRp': [True, None],
                'sigma11': [True, None],
                'sigma12': [True, None],
                'sigma13': [True, None],
                'sigma14': [True, None],
                'sigma15': [True, None],
                'sigma16': [True, None],
                'sigma22': [True, None],
                'sigma23': [True, None],
                'sigma24': [True, None],
                'sigma25': [True, None],
                'sigma26': [True, None],
                'sigma33': [True, None],
                'sigma34': [True, None],
                'sigma35': [True, None],
                'sigma36': [True, None],
                'sigma44': [True, None],
                'sigma45': [True, None],
                'sigma46': [True, None],
                'sigma55': [True, None],
                'sigma56': [True, None],
                'sigma66': [True, None],
                'shellX': [True, None],
                'shellXp': [True, None],
                'shellY': [True, None],
                'shellYp': [True, None],
                'shellXWidth': [True, None],
                'shellXpWidth': [True, None],
                'shellYWidth': [True, None],
                'shellYpWidth': [True, None],
                'Rmin': [True, None],
                'Rmax': [True, None],
                'haloNSigmaXInner': [True, None],
                'haloNSigmaXOuter': [True, None],
                'haloNSigmaYInner': [True, None],
                'haloNSigmaYOuter': [True, None],
                'haloXCutInner': [True, None],
                'haloYCutInner': [True, None],
                'haloPSWeightParameter': [True, None],
                'haloPSWeightFunction': [True, None],
                'offsetSampleMean': [True, None],
                'eventGeneratorMinX': [True, None],
                'eventGeneratorMaxX': [True, None],
                'eventGeneratorMinY': [True, None],
                'eventGeneratorMaxY': [True, None],
                'eventGeneratorMinZ': [True, None],
                'eventGeneratorMaxZ': [True, None],
                'eventGeneratorMinXp': [True, None],
                'eventGeneratorMaxXp': [True, None],
                'eventGeneratorMinYp': [True, None],
                'eventGeneratorMaxYp': [True, None],
                'eventGeneratorMinZp': [True, None],
                'eventGeneratorMaxZp': [True, None],
                'eventGeneratorMinT': [True, None],
                'eventGeneratorMaxT': [True, None],
                'eventGeneratorMinEK': [True, None],
                'eventGeneratorMaxEK': [True, None],
                'eventGeneratorParticles': [True, None],
            }

    class Geant4Data(Output.Tree):
        # https://github.com/scikit-hep/uproot/issues/468
        ...

    class Options(Output.Tree):
        class OptionsBase(Output.Branch):
            BRANCH_NAME = 'GMAD::OptionsBase'
            DEFAULT_LEAVES = {
                'inputFileName': [True, None],
                'visMacroFileName': [True, None],
                'geant4MacroFileName': [True, None],
                'visDebug': [True, None],
                'outputFileName': [True, None],
                'outputFormat': [True, None],
                'outputDoublePrecision': [True, None],
                'survey': [True, None],
                'surveyFileName': [True, None],
                'batch': [True, None],
                'verbose': [True, None],
                'verboseRunLevel': [True, None],
                'verboseEventBDSIM': [True, None],
                'verboseEventLevel': [True, None],
                'verboseEventStart': [True, None],
                'verboseEventContinueFor': [True, None],
                'verboseTrackingLevel': [True, None],
                'verboseSteppingBDSIM': [True, None],
                'verboseSteppingLevel': [True, None],
                'verboseSteppingEventStart': [True, None],
                'verboseSteppingEventContinueFor': [True, None],
                'verboseSteppingPrimaryOnly': [True, None],
                'verboseImportanceSampling': [True, None],
                'circular': [True, None],
                'seed': [True, None],
                'nGenerate': [True, None],
                'recreate': [True, None],
                'recreateFileName': [True, None],
                'startFromEvent': [True, None],
                'writeSeedState': [True, None],
                'useASCIISeedState': [True, None],
                'seedStateFileName': [True, None],
                'generatePrimariesOnly': [True, None],
                'exportGeometry': [True, None],
                'exportType': [True, None],
                'exportFileName': [True, None],
                'bdsimPath': [True, None],
                'physicsList': [True, None],
                'physicsVerbose': [True, None],
                'physicsVerbosity': [True, None],
                'physicsEnergyLimitLow': [True, None],
                'physicsEnergyLimitHigh': [True, None],
                'g4PhysicsUseBDSIMRangeCuts': [True, None],
                'g4PhysicsUseBDSIMCutsAndLimits': [True, None],
                'eventOffset': [True, None],
                'recreateSeedState': [True, None],
                'elossHistoBinWidth': [True, None],
                'ffact': [True, None],
                'beamlineX': [True, None],
                'beamlineY': [True, None],
                'beamlineZ': [True, None],
                'beamlinePhi': [True, None],
                'beamlineTheta': [True, None],
                'beamlinePsi': [True, None],
                'beamlineAxisX': [True, None],
                'beamlineAxisY': [True, None],
                'beamlineAxisZ': [True, None],
                'beamlineAngle': [True, None],
                'beamlineAxisAngle': [True, None],
                'beamlineS': [True, None],
                'eventNumberOffset': [True, None],
                'checkOverlaps': [True, None],
                'xsize': [True, None],
                'ysize': [True, None],
                'magnetGeometryType': [True, None],
                'outerMaterialName': [True, None],
                'horizontalWidth': [True, None],
                'thinElementLength': [True, None],
                'hStyle': [True, None],
                'vhRatio': [True, None],
                'coilWidthFraction': [True, None],
                'coilHeightFraction': [True, None],
                'ignoreLocalMagnetGeometry': [True, None],
                'preprocessGDML': [True, None],
                'preprocessGDMLSchema': [True, None],
                'dontSplitSBends': [True, None],
                'yokeFields': [True, None],
                'includeFringeFields': [True, None],
                'includeFringeFieldsCavities': [True, None],
                'beampipeThickness': [True, None],
                'apertureType': [True, None],
                'aper1': [True, None],
                'aper2': [True, None],
                'aper3': [True, None],
                'aper4': [True, None],
                'beampipeMaterial': [True, None],
                'ignoreLocalAperture': [True, None],
                'vacMaterial': [True, None],
                'emptyMaterial': [True, None],
                'worldMaterial': [True, None],
                'worldGeometryFile': [True, None],
                'autoColourWorldGeometryFile': [True, None],
                'importanceWorldGeometryFile': [True, None],
                'importanceVolumeMap': [True, None],
                'worldVolumeMargin': [True, None],
                'vacuumPressure': [True, None],
                'buildTunnel': [True, None],
                'buildTunnelStraight': [True, None],
                'tunnelType': [True, None],
                'tunnelThickness': [True, None],
                'tunnelSoilThickness': [True, None],
                'tunnelMaterial': [True, None],
                'soilMaterial': [True, None],
                'buildTunnelFloor': [True, None],
                'tunnelFloorOffset': [True, None],
                'tunnelAper1': [True, None],
                'tunnelAper2': [True, None],
                'tunnelVisible': [True, None],
                'tunnelOffsetX': [True, None],
                'tunnelOffsetY': [True, None],
                'removeTemporaryFiles': [True, None],
                'samplerDiameter': [True, None],
                'turnOnOpticalAbsorption': [True, None],
                'turnOnMieScattering': [True, None],
                'turnOnRayleighScattering': [True, None],
                'turnOnOpticalSurface': [True, None],
                'scintYieldFactor': [True, None],
                'maximumPhotonsPerStep': [True, None],
                'maximumBetaChangePerStep': [True, None],
                'maximumTracksPerEvent': [True, None],
                'minimumKineticEnergy': [True, None],
                'minimumKineticEnergyTunnel': [True, None],
                'minimumRange': [True, None],
                'defaultRangeCut': [True, None],
                'prodCutPhotons': [True, None],
                'prodCutElectrons': [True, None],
                'prodCutPositrons': [True, None],
                'prodCutProtons': [True, None],
                'neutronTimeLimit': [True, None],
                'neutronKineticEnergyLimit': [True, None],
                'useLENDGammaNuclear': [True, None],
                'useElectroNuclear': [True, None],
                'useMuonNuclear': [True, None],
                'useGammaToMuMu': [True, None],
                'usePositronToMuMu': [True, None],
                'usePositronToHadrons': [True, None],
                'collimatorsAreInfiniteAbsorbers': [True, None],
                'tunnelIsInfiniteAbsorber': [True, None],
                'defaultBiasVacuum': [True, None],
                'defaultBiasMaterial': [True, None],
                'integratorSet': [True, None],
                'lengthSafety': [True, None],
                'lengthSafetyLarge': [True, None],
                'maximumTrackingTime': [True, None],
                'maximumStepLength': [True, None],
                'maximumTrackLength': [True, None],
                'chordStepMinimum': [True, None],
                'chordStepMinimumYoke': [True, None],
                'deltaIntersection': [True, None],
                'minimumEpsilonStep': [True, None],
                'maximumEpsilonStep': [True, None],
                'deltaOneStep': [True, None],
                'stopSecondaries': [True, None],
                'killNeutrinos': [True, None],
                'minimumRadiusOfCurvature': [True, None],
                'sampleElementsWithPoleface': [True, None],
                'nominalMatrixRelativeMomCut': [True, None],
                'teleporterFullTransform': [True, None],
                'sensitiveOuter': [True, None],
                'sensitiveBeamPipe': [True, None],
                'numberOfEventsPerNtuple': [True, None],
                'storeApertureImpacts': [True, None],
                'storeApertureImpactsIons': [True, None],
                'storeApertureImpactsAll': [True, None],
                'apertureImpactsMinimumKE': [True, None],
                'storeCollimatorInfo': [True, None],
                'storeCollimatorHits': [True, None],
                'storeCollimatorHitsLinks': [True, None],
                'storeCollimatorHitsIons': [True, None],
                'storeCollimatorHitsAll': [True, None],
                'collimatorHitsMinimumKE': [True, None],
                'storeEloss': [True, None],
                'storeElossHistograms': [True, None],
                'storeElossVacuum': [True, None],
                'storeElossVacuumHistograms': [True, None],
                'storeElossTunnel': [True, None],
                'storeElossTunnelHistograms': [True, None],
                'storeElossWorld': [True, None],
                'storeElossWorldContents': [True, None],
                'storeElossTurn': [True, None],
                'storeElossLinks': [True, None],
                'storeElossLocal': [True, None],
                'storeElossGlobal': [True, None],
                'storeElossTime': [True, None],
                'storeElossStepLength': [True, None],
                'storeElossPreStepKineticEnergy': [True, None],
                'storeElossModelID': [True, None],
                'storeTrajectory': [True, None],
                'storeTrajectoryDepth': [True, None],
                'storeTrajectoryParticle': [True, None],
                'storeTrajectoryParticleID': [True, None],
                'storeTrajectoryEnergyThreshold': [True, None],
                'storeTrajectorySamplerID': [True, None],
                'storeTrajectoryELossSRange': [True, None],
                'storeTrajectoryTransportationSteps': [True, None],
                'trajNoTransportation': [True, None],
                'storeTrajectoryLocal': [True, None],
                'storeTrajectoryLinks': [True, None],
                'storeTrajectoryIon': [True, None],
                'trajectoryFilterLogicAND': [True, None],
                'storeSamplerAll': [True, None],
                'storeSamplerPolarCoords': [True, None],
                'storeSamplerCharge': [True, None],
                'storeSamplerKineticEnergy': [True, None],
                'storeSamplerMass': [True, None],
                'storeSamplerRigidity': [True, None],
                'storeSamplerIon': [True, None],
                'trajCutGTZ': [True, None],
                'trajCutLTR': [True, None],
                'trajConnect': [True, None],
                'storeModel': [True, None],
                'nturns': [True, None],
                'ptcOneTurnMapFileName': [True, None],
                'printFractionEvents': [True, None],
                'printFractionTurns': [True, None],
                'printPhysicsProcesses': [True, None],
                'nSegmentsPerCircle': [True, None],
                'nbinsx': [True, None],
                'nbinsy': [True, None],
                'nbinsz': [True, None],
                'xmin': [True, None],
                'xmax': [True, None],
                'ymin': [True, None],
                'ymax': [True, None],
                'zmin': [True, None],
                'zmax': [True, None],
                'useScoringMap': [True, None],
            }

    class Model(Output.Tree):
        @property
        def component_names(self):
            return self.model.component_names

        @property
        def placement_names(self):
            return self.model.placement_names

        @property
        def sampler_names(self):
            return self.model.sampler_names

        @property
        def collimator_names(self):
            return self.model.collimator_names

        @property
        def scoring_mesh_names(self):
            return self.model.scoring_mesh_names

        @property
        def scoring_mesh_translations(self):
            return self.model.scoring_mesh_translations

        @property
        def scoring_mesh_rotations(self):
            return self.model.scoring_mesh_rotations

        class Model(Output.Branch):

            DEFAULT_LEAVES = {
                'n': [True, None],
                'samplerNamesUnique': [False, None],
                'componentName': [False, None],
                'placementName': [False, None],
                'componentType': [True, None],
                'length': [True, None],
                'angle': [True, None],
                'staPos': [True, None],
                'midPos': [True, None],
                'endPos': [True, None],
                'staRot': [False, None],
                'midRot': [False, None],
                'endRot': [False, None],
                'staRefPos': [True, None],
                'midRefPos': [True, None],
                'endRefPos': [True, None],
                'staRefRot': [True, None],
                'midRefRot': [True, None],
                'endRefRot': [True, None],
                'tilt': [True, None],
                'offsetX': [True, None],
                'offsetY': [True, None],
                'staS': [True, None],
                'midS': [True, None],
                'endS': [True, None],
                'beamPipeType': [True, None],
                'beamPipeAper1': [True, None],
                'beamPipeAper2': [True, None],
                'beamPipeAper3': [True, None],
                'beamPipeAper4': [True, None],
                'material': [True, None],
                'k1': [True, None],
                'k2': [True, None],
                'k3': [True, None],
                'k4': [False, None],
                'k5': [False, None],
                'k6': [False, None],
                'k7': [False, None],
                'k8': [False, None],
                'k9': [False, None],
                'k10': [False, None],
                'k11': [False, None],
                'k12': [False, None],
                'k1s': [False, None],
                'k2s': [False, None],
                'k3s': [False, None],
                'k4s': [False, None],
                'k5s': [False, None],
                'k6s': [False, None],
                'k7s': [False, None],
                'k8s': [False, None],
                'k9s': [False, None],
                'k10s': [False, None],
                'k11s': [False, None],
                'k12s': [False, None],
                'ks': [False, None],
                'hkick': [True, None],
                'vkick': [True, None],
                'bField': [True, None],
                'eField': [True, None],
                'e1': [True, None],
                'e2': [True, None],
                'hgap': [True, None],
                'fint': [True, None],
                'fintx': [True, None],
                'fintk2': [True, None],
                'fintxk2': [True, None],
                'storeCollimatorInfo': [False, None],
                'collimatorIndices': [False, None],
                'collimatorIndicesByName': [False, None],
                'nCollimators': [False, None],
                'collimatorInfo': [False, None],
                'collimatorBranchNamesUnique': [False, None],
                'scoringMeshName': [False, None],
                'scoringMeshRotation': [False, None],
                'scoringMeshTranslation': [False, None]
            }

            @property
            def component_names(self):
                return self.array('componentName')

            @property
            def placement_names(self):
                return self.array('placementName')

            @property
            def sampler_names(self):
                return self.array('samplerNamesUnique')

            @property
            def collimator_names(self):
                return self.array('collimatorBranchNamesUnique')

            @property
            def scoring_mesh_names(self):
                return self.array('scoringMeshName')

            @property
            def scoring_mesh_translations(self):
                return dict(tuple(self.array('scoringMeshTranslation')))

            @property
            def scoring_mesh_rotations(self):
                return dict(tuple(self.array('scoringMeshRotation')))

        def to_df(self) -> _pd.DataFrame:
            """

            Returns:

            """
            model_geometry_df = _pd.DataFrame()

            # Names and strings
            for branch, name in {'componentName': 'NAME',
                                 'componentType': 'TYPE',
                                 'material': 'MATERIAL',
                                 'beamPipeType': 'APERTYPE',
                                 }.items():
                data = self.array(branch=branch)
                model_geometry_df[name] = data

            # Scalar
            for branch, name in {'length': 'L',
                                 'angle': 'ANGLE',
                                 'staS': 'AT_ENTRY',
                                 'midS': 'AT_CENTER',
                                 'endS': 'AT_EXIT',
                                 'tilt': 'TILT',
                                 'k1': 'K1',
                                 'k2': 'K2',
                                 'k3': 'K3',
                                 'k4': 'K4',
                                 'k5': 'K5',
                                 'k6': 'K6',
                                 'k7': 'K7',
                                 'k8': 'K8',
                                 'k9': 'K9',
                                 'k10': 'K10',
                                 'k11': 'K11',
                                 'k12': 'K12',
                                 'k1s': 'K1S',
                                 'k2s': 'K2S',
                                 'k3s': 'K3S',
                                 'k4s': 'K4S',
                                 'k5s': 'K5S',
                                 'k6s': 'K6S',
                                 'k7s': 'K7S',
                                 'k8s': 'K8S',
                                 'k9s': 'K9S',
                                 'k10s': 'K10S',
                                 'k11s': 'K11S',
                                 'k12s': 'K12S',
                                 'bField': 'B',
                                 'eField': 'E',
                                 'e1': 'E1',
                                 'e2': 'E2',
                                 'hgap': 'HGAP',
                                 'fint': 'FINT',
                                 'fintx': 'FINTX'
                                 }.items():
                model_geometry_df[name] = self.array(branch=branch)

            # Aperture
            for branch, name in {'beamPipeAper1': 'APERTURE1',
                                 'beamPipeAper2': 'APERTURE2',
                                 'beamPipeAper3': 'APERTURE3',
                                 'beamPipeAper4': 'APERTURE4'}.items():
                model_geometry_df[name] = self.array(branch=branch)

            # Collimators
            if not self.array(branch='storeCollimatorInfo'):
                _logging.warning("Informations for collimators are empty, you shoud set storeCollimatorInfo=1")

            collimator_df = _pd.DataFrame()
            branches = ['collimatorInfo/Model.collimatorInfo.componentName',
                        'collimatorInfo/Model.collimatorInfo.xSizeIn',
                        'collimatorInfo/Model.collimatorInfo.ySizeIn']
            collimator_df = _pd.DataFrame(self.arrays(branches)).T
            collimator_df.columns=['NAME', 'APERTURE1', 'APERTURE2']

            for _, line in collimator_df.iterrows():
                col_name = line['NAME']
                idx = model_geometry_df.query(f"NAME == @col_name").index.values[0]
                model_geometry_df.at[idx, 'APERTURE1'] = line['APERTURE1']
                model_geometry_df.at[idx, 'APERTURE2'] = line['APERTURE2']

            # Vectors
            geometry_branches = {'staPos': 'ENTRY_',
                                 'midPos': 'CENTER_',
                                 'endPos': 'EXIT_'}

            data_df = _pd.DataFrame()
            for branch, name in geometry_branches.items():
                data = _pd.DataFrame(self.array(branch)).rename({"fX": f"{name}X", "fY": f"{name}Y", "fZ": f"{name}Z"},
                                                                axis='columns')
                data_df = _pd.concat([data_df, data], axis='columns')

            # Concatenate
            self._df = _pd.concat([model_geometry_df, data_df], axis='columns', sort=False).set_index('NAME')

            return self._df

    class Run(Output.Tree):

        class Summary(Output.Branch):
            DEFAULT_LEAVES = {
                'startTime': [True, None],
                'stopTime': [True, None],
                'durationWall': [True, None],
                'durationCPU': [True, None],
                'seedStateAtStart': [True, None],
            }

        class Histos(Output.Branch):
            DEFAULT_LEAVES = {

            }

    class Event(Output.Tree):
        def __getattr__(self, item):
            if item == 'samplers':
                self.samplers = BDSIMOutput.Event.Samplers({
                    s.rstrip('.'):
                        BDSIMOutput.Event.Sampler(parent=self, branch_name=s)
                    for s in self.parent.model.sampler_names})
                return self.samplers

            elif item == 'collimators':
                self.collimators = BDSIMOutput.Event.Collimators({
                    s.rstrip('.'):
                        BDSIMOutput.Event.Collimator(parent=self, branch_name=s)
                    for s in self.parent.model.collimator_names})
                return self.collimators
            else:
                return super().__getattr__(item)

        class Eloss(Output.Branch):
            DEFAULT_LEAVES = {
                'n': [True, None],
                'energy': [True, None],
                'S': [True, None],
                'weight': [True, None],
                'partID': [False, None],
                'trackID': [False, None],
                'parentID': [False, None],
                'modelID': [False, None],
                'turn': [False, None],
                'x': [False, None],
                'y': [False, None],
                'z': [False, None],
                'X': [False, None],
                'Y': [False, None],
                'Z': [False, None],
                'T': [False, None],
                'stepLength': [False, None],
                'preStepKineticEnergy': [False, None],
                'storeTurn': [True, None],
                'storeLinks': [True, None],
                'storeModelID': [True, None],
                'storeLocal': [True, None],
                'storeGlobal': [True, None],
                'storeTime': [True, None],
                'storeStepLength': [True, None],
                'storePreStepKineticEnergy': [True, None],
            }

        class ELossVacuum(Eloss):
            pass

        class ELossTunnel(Eloss):
            pass

        class ELossWorld(Eloss):
            pass

        class ELossWorldExit(Eloss):
            pass

        class Primary(Output.Branch):
            DEFAULT_LEAVES = {
                'energy': [True, None],
                'weight': [True, None],
                'partID': [False, None],
                'trackID': [False, None],
                'parentID': [False, None],
                'modelID': [False, None],
                'turnNumber': [True, None],
                'x': [True, None],
                'y': [True, None],
                'z': [True, None],
                'xp': [True, None],
                'yp': [True, None],
                'zp': [True, None],
                'p': [True, None],
                'T': [True, None],
                'S': [False, None],
                'r': [False, None],
                'n': [False, None],
                'rp': [False, None],
                'phi': [False, None],
                'phip': [False, None],
                'theta': [False, None],
                'charge': [False, None],
                'kineticEnergy': [False, None],
                'mass': [False, None],
                'rigidity': [False, None],
                'isIon': [False, None],
                'ionA': [False, None],
                'ionZ': [False, None],
                'nElectrons': [False, None],
            }

        class PrimaryFirstHit(Output.Branch):
            DEFAULT_LEAVES = {
                'S': [True, None],
                'energy': [True, None],
                'weight': [True, None],
                'partID': [False, None],
                'trackID': [False, None],
                'parentID': [False, None],
                'modelID': [False, None],
                'turn': [True, None],
                'x': [True, None],
                'y': [True, None],
                'z': [True, None],
                'X': [True, None],
                'Y': [True, None],
                'Z': [True, None],
                'T': [True, None],
                'stepLength': [False, None],
                'preStepKineticEnergy': [False, None],
                'n': [False, None],
                'storeTurn': [False, None],
                'storeLinks': [False, None],
                'storeModelID': [False, None],
                'storeLocal': [False, None],
                'storeGlobal': [False, None],
                'storeTime': [False, None],
                'storeStepLength': [False, None],
                'storePreStepKineticEnergy': [False, None],
            }

        class PrimaryLastHit(PrimaryFirstHit):
            pass

        class ApertureImpacts(Output.Branch):
            DEFAULT_LEAVES = {
                'n': [True, None],
                'energy': [True, None],
                'S': [True, None],
                'weight': [True, None],
                'isPrimary': [True, None],
                'firstPrimaryImpact': [True, None],
                'partID': [True, None],
                'turn': [False, None],
                'x': [False, None],
                'y': [False, None],
                'xp': [False, None],
                'yp': [False, None],
                'T': [False, None],
                'kineticEnergy': [False, None],
                'isIon': [False, None],
                'ionA': [False, None],
                'ionZ': [False, None],
                'trackID': [False, None],
                'parentID': [False, None],
                'modelID': [False, None],
            }

        class Histos(Output.Branch):
            def read_df(self) -> _pd.DataFrame:
                pass

        class Samplers(_UserDict):
            def compute_optics(self, samplers: _Optional[_List[str]] = None):
                return _pd.DataFrame(
                    [sampler.compute_optics() for sampler in self.data.values()]
                )

            def to_df(self, samplers: _Optional[_List[str]] = None, columns: _Optional[_List[str]] = None) -> _pd.DataFrame:
                pass

            def to_np(self, samplers: _Optional[_List[str]] = None, columns: _Optional[_List[str]] = None) -> _np.ndarray:
                pass

            @property
            def df(self) -> _pd.DataFrame:
                return self.to_df()

            @property
            def np(self) -> _np.ndarray:
                return self.to_np()

            @property
            def optics(self):
                if self._optics is None:
                    self._optics = self.compute_optics()
                return self._optics

        class Sampler(Output.Branch):
            DEFAULT_LEAVES = {
                'x': [True, None],
                'xp': [True, None],
                'y': [True, None],
                'yp': [True, None],
                'z': [True, None],
                'zp': [True, None],
                'T': [True, None],
                'energy': [True, None],
                'p': [True, None],
                'kineticEnergy': [True, None],
                'turnNumber': [True, None],
                'parentID': [True, None],
                'partID': [True, None],
                'trackID': [True, None],
                'weight': [True, None],
                'n': [True, None],
                'S': [True, None],
            }

            def to_np(self,
                      turn_number: int = -1,
                      primary_only: bool = True,
                      ) -> _np.ndarray:
                df: _pd.DataFrame = self.to_df()
                data = df[['x', 'xp', 'y', 'yp', 'T', 'energy', 'n', 'S']].values
                validity = df[['turnNumber', 'parentID']].values
                if turn_number == - 1 and primary_only is False:
                    return data
                elif turn_number == -1 and primary_only is True:
                    return data[validity[:, 1] == 0]
                elif primary_only is False:
                    return data[validity[:, 0] == turn_number]
                else:
                    return data[_np.logical_and(validity[:, 1] == 0, validity[:, 0] == turn_number), :]

            def compute_optics(self):
                """

                Returns:

                """
                data = self.to_np(turn_number=1, primary_only=True)
                cv = _np.cov(data)
                eps_x = _np.sqrt(cv[0, 0] * cv[1, 1] - cv[0, 1] * cv[1, 0])
                eps_y = _np.sqrt(cv[2, 2] * cv[3, 3] - cv[2, 3] * cv[3, 2])
                return {
                    'BETA11': (cv[0, 0] - cv[0, 5] ** 2) / eps_x,
                    'BETA22': (cv[2, 2] - cv[2, 5] ** 2) / eps_y,
                    'ALPHA11': -cv[1, 1] / eps_x,
                    'ALPHA22': -cv[3, 3] / eps_y,
                    'DISP1': cv[0, 5] / 0.001,
                    'DISP2': 0.0,
                    'DISP3': cv[2, 5] / 0.001,
                    'DISP4': 0.0,
                    'EPSX': eps_x,
                    'EPXY': eps_y,
                    'n': data[:, -2].sum(),
                    'S': data[0, -1],
                }

        class Collimators(_UserDict):
            def to_df(self, samplers: _Optional[_List[str]] = None, columns: _Optional[_List[str]] = None) -> _pd.DataFrame:
                ...

            def to_np(self, samplers: _Optional[_List[str]] = None, columns: _Optional[_List[str]] = None) -> _np.ndarray:
                ...

            @property
            def df(self) -> _pd.DataFrame:
                return self.to_df()

            @property
            def np(self) -> _np.ndarray:
                return self.to_np()

        class Collimator(Output.Branch):
            DEFAULT_LEAVES = {
                'primaryInteracted': [True, None],
                'primaryStopped': [True, None],
                'n': [True, None],
                'energy': [False, None],
                'energyDeposited': [False, None],
                'xln': [False, None],
                'yln': [False, None],
                'zln': [False, None],
                'xpln': [False, None],
                'ypln': [False, None],
                'T': [False, None],
                'weight': [False, None],
                'partID': [False, None],
                'parentID': [False, None],
                'turn': [False, None],
                'firstPrimaryHitThisTurn': [False, None],
                'impactParameterX': [False, None],
                'impactParameterY': [False, None],
                'isIon': [False, None],
                'ionA': [False, None],
                'ionZ': [False, None],
                'turnSet': [False, None],
                'charge': [False, None],
                'kineticEnergy': [False, None],
                'mass': [False, None],
                'rigidity': [False, None],
                'totalEnergyDeposited': [False, None],
            }


class REBDSIMOutput(Output):
    def __getattr__(self, item):
        if item in (
                'beam',
                'event',
                'run',
                'options',
                'model_dir'
        ):
            try:
                self._root_directory.get(item.split('_')[0].title())
            except KeyError:
                raise BDSIMOutputException(f"Key {item} is invalid.")
            setattr(self,
                    item,
                    Output.Directory(parent=self, directory=self._root_directory[item.split('_')[0].title()])
                    )
        elif item == 'model':
            setattr(self,
                    item,
                    getattr(BDSIMOutput, item.title())(parent=self, tree_name='ModelTree'),
                    )
        else:
            setattr(self,
                    item,
                    getattr(BDSIMOutput, item.title())(parent=self)
                    )
        return getattr(self, item)


class REBDSIMOpticsOutput(Output):
    def __getattr__(self, item):
        try:
            self._root_directory.get(item.title())
        except KeyError:
            raise BDSIMOutputException(f"Key {item} is invalid.")

        if item in (
                'optics',
        ):
            setattr(self,
                    item,
                    getattr(REBDSIMOpticsOutput, item.title())(parent=self)
                    )
            return getattr(self, item)
        else:
            setattr(self,
                    item,
                    getattr(BDSIMOutput, item.title())(parent=self)
                    )
        return getattr(self, item)

    class Optics(Output.Tree):

        class Optics(Output.Branch):
            BRANCH_NAME = ''
            DEFAULT_LEAVES = {
                'Emitt_x': [True, None],
                'Emitt_y': [True, None],
                'Alpha_x': [True, None],
                'Alpha_y': [True, None],
                'Beta_x': [True, None],
                'Beta_y': [True, None],
                'Gamma_x': [True, None],
                'Gamma_y': [True, None],
                'Disp_x': [True, None],
                'Disp_y': [True, None],
                'Disp_xp': [True, None],
                'Disp_yp': [True, None],
                'Mean_x': [True, None],
                'Mean_y': [True, None],
                'Mean_xp': [True, None],
                'Mean_yp': [True, None],
                'Sigma_x': [True, None],
                'Sigma_y': [True, None],
                'Sigma_xp': [True, None],
                'Sigma_yp': [True, None],
                'S': [True, None],
                'Npart': [True, None],
                'Sigma_Emitt_x': [True, None],
                'Sigma_Emitt_y': [True, None],
                'Sigma_Alpha_x': [True, None],
                'Sigma_Alpha_y': [True, None],
                'Sigma_Beta_x': [True, None],
                'Sigma_Beta_y': [True, None],
                'Sigma_Gamma_x': [True, None],
                'Sigma_Gamma_y': [True, None],
                'Sigma_Disp_x': [True, None],
                'Sigma_Disp_y': [True, None],
                'Sigma_Disp_xp': [True, None],
                'Sigma_Disp_yp': [True, None],
                'Sigma_Mean_x': [True, None],
                'Sigma_Mean_y': [True, None],
                'Sigma_Mean_xp': [True, None],
                'Sigma_Mean_yp': [True, None],
                'Sigma_Sigma_x': [True, None],
                'Sigma_Sigma_y': [True, None],
                'Sigma_Sigma_xp': [True, None],
                'Sigma_Sigma_yp': [True, None],
                'Mean_E': [True, None],
                'Mean_t': [True, None],
                'Sigma_E': [True, None],
                'Sigma_t': [True, None],
                'Sigma_Mean_E': [True, None],
                'Sigma_Mean_t': [True, None],
                'Sigma_Sigma_E': [True, None],
                'Sigma_Sigma_t': [True, None],
                'xyCorrelationCoefficent': [True, None],
            }

        # Override the method to_df as rebsdimOptics as not the same structure
        def to_df(self) -> _pd.DataFrame:
            """

            Returns:

            """
            optics_df = _pd.DataFrame(self.arrays(branches=self.Optics.DEFAULT_LEAVES.keys())).T
            optics_df.columns = self.Optics.DEFAULT_LEAVES.keys()
            return optics_df


class REBDSIMCombineOutput(REBDSIMOutput):
    def __getattr__(self, item):
        if item not in ('event', 'header'):
            return None
        return super().__getattr__(item)



import ROOT as _ROOT

class Model :
    def __init__(self, filename):
        self._filename = filename
        self._rootFile = _ROOT.TFile(filename)
        self._tree     = self._rootFile.Get("Model")
        self._entry = _ROOT.BDSOutputROOTEventModel()
        self._tree.SetBranchAddress("Model.",self._entry)
        self._tree.GetEntry(0)




import ROOT             as _ROOT
from ROOT import TChain as _TChain
from ROOT import TH1F   as _TH1F
from ROOT import TFile  as _TFile

import matplotlib.pyplot as _plt

class AnalysisRoot:
    def __init__(self,filelist):
        if filelist != None:
            self.Load(filelist)

    def Load(self,filelist):
        self.filelist = filelist
        self.nfiles   = 0
        self._LoadPrimaries()
        self._LoadElossTree()
        self._LoadPlossTree()
        self._LoadSamplerTree()
        if self.nfiles == 0:
            self._NoFilesWarning()

    def _LoadPrimaries(self) : 
        self.primaries = _TChain("primaries") 
        if type(self.filelist) == list : 
            for f in self.filelist : 
                self.primaries.Add(f)
                
    def _LoadSamplerTree(self) : 
        # open a single file and determine the available samplers 
        f = _TFile.Open(self.filelist[0])
        k = f.GetListOfKeys()

        self.samplerNames = [] 
        self.samplerDict  = {}        

        for i in  range(0,k.GetEntries(),1) : 
            if k[i].GetName().find("Sampler") == 0 : 
                self.samplerNames.append(k[i].GetName()) 

        # Sampler chain dictionary
        for n in self.samplerNames : 
            self.samplerDict[n] = _TChain(n)
            for f in self.filelist : 
                self.samplerDict[n].Add(f)

    def _LoadElossTree(self):
        self.elossch = _TChain("ElossTree")
        if type(self.filelist) == list:
            for f in self.filelist:
                self.elossch.Add(f)
                self.nfiles += 1
        else:
            self.nfiles = self.elossch.Add(self.filelist)
        
    def _LoadPlossTree(self):
        self.plossch = _TChain("PlossTree")
        if type(self.filelist) == list:
            for f in self.filelist:
                self.plossch.Add(f)
                self.nfiles += 1
        else:
            self.nfiles = self.plossch.Add(self.filelist)

    def _NoFilesWarning(self):
        raise UserWarning("No root files were loaded")
        
    def PlotEnergyLossHistogram(self,nbins=100,xmin=0,xmax=100,normalised=False,weighted=True):
        self.elossch.SetBranchStatus("s",1)
        if weighted:
            self.elossch.SetBranchStatus("E",1)
        self.elosshist = _TH1F("hist","Energy Loss",nbins,xmin,xmax)

        #fill histogram
        nentries = int(self.elossch.GetEntries())
        for i in range(nentries):
            self.elossch.GetEntry(i)
            if weighted:
                self.elosshist.Fill(self.elossch.s,self.elossch.E)
            else:
                self.elosshist.Fill(self.elossch.s)
        
        #normalise if required
        if normalised:
            scale = 1./max(self.elosshist)
            print('Normalising by factor',scale)
            self.normalisation = scale
            self.elosshist.Scale(scale)
        
        _PlotHistogram(self.elosshist,normalised)

    def PlotPrimaryLossHistogram(self,nbins=100,xmin=0,xmax=100,normalised=True,weighted=False):
        self.plossch.SetBranchStatus("s",1)
        if weighted:
            self.plossch.SetBranchStatus("E",1)
        self.plosshist = _TH1F("hist","Primary Losses",nbins,xmin,xmax)

        #fill histogram
        nentries = int(self.plossch.GetEntries())
        for i in range(nentries):
            self.plossch.GetEntry(i)
            if weighted:
                self.plosshist.Fill(self.plossch.s,self.plossch.E)
            else:
                self.plosshist.Fill(self.plossch.s)
        
        #normalise if required
        if normalised:
            scale = 1./max(self.plosshist)
            self.normalisation = scale
            print('Normalising by factor',scale)
            self.plosshist.Scale(scale)
        
        _PlotHistogram(self.plosshist,normalised)
        

def _PlotHistogram(hist,normalised,title=''):
    #plot histogram
    fig = _plt.figure(figsize=(12,5))
    ax  = fig.add_subplot(111)
    ax.plot(hist)
    #set x limits to add 2% on either side
    xr = len(hist)
    ax.set_xlim(0-0.02*xr,xr+0.02*xr)
    #set large log yscale
    if not normalised:
        ax.set_ylim(min(hist)*0.7,max(hist*1.4))
    else:
        ax.set_ylim(0.7e-7,1.4)
    
    ax.set_xlabel('S Position (m)')
    if not normalised:
        ax.set_ylabel('Number')
    else:
        ax.set_ylabel('Fraction')
    ax.set_yscale('log')
    ax.grid()
    ax.set_title(title)
    _plt.subplots_adjust(left=0.08,right=0.97)


#! /usr/bin/env python2.7

import pybdsim

a = pybdsim.Builder.Machine()

a.AddDrift('d1',1.2,apertureType="rectangular", aper1=(3,"cm"), aper2=(1.5,"cm"))
a.AddQuadrupole("q1",0.2,k1=0.034515)
a.AddDrift('d2',0.5)
a.AddQuadrupole("q2",0.2,k1=-0.034515)
a.AddDrift('d3',0.76)
a.AddDipole('dp','sbend',4.5,0.0789)
a.AddDrift('d4',0.6)

a.AddSampler("all")

a.Write("testlattice")



import pybdsim as _pybdsim
import matplotlib.pyplot as _plt
from os.path import isfile as _isfile
from matplotlib.backends.backend_pdf import PdfPages as _PdfPages
import datetime as _datetime

from pybdsim.Data import LoadSDDSColumnsToDict as _LoadSDDSColumnsToDict
from pybdsim.Plot import AddMachineLatticeFromSurveyToFigure as _AddMachineLatticeFromSurveyToFigure

def ElegantVsBDSIM(elegantTwiss, elegantSigma, elegantCentroid, bdsim, functions=None,
                   postfunctions=None, figsize=(10, 5), saveAll=True, outputFileName=None,
                   particleType="e-"):
    """
    Compares MadX and BDSIM optics variables.
    User must provide a tfsoptIn file or Tfsinstance and a BDSAscii file or instance.

    +-----------------+---------------------------------------------------------+
    | **Parameters**  | **Description**                                         |
    +-----------------+---------------------------------------------------------+
    | twiss           | Elegan twiss file instance.                             |
    +-----------------+---------------------------------------------------------+
    | bdsim           | Optics root file (from rebdsimOptics or rebdsim).       |
    +-----------------+---------------------------------------------------------+
    | survey          | BDSIM model survey.                                     |
    +-----------------+---------------------------------------------------------+
    | functions       | Hook for users to add their functions that are called   |
    |                 | immediately prior to the addition of the plot. Use a    |
    |                 | lambda function to add functions with arguments. Can    |
    |                 | be a function or a list of functions.                   |
    +-----------------+---------------------------------------------------------+
    | figsize         | Figure size for all figures - default is (12,5)         |
    +-----------------+---------------------------------------------------------+
    
    """

    _CheckFileExistsList(elegantTwiss, elegantSigma, elegantCentroid, bdsim)

    fname = _pybdsim.Data.GetFileName(bdsim) # cache file name
    if fname == "":
        fname = "optics_report"

    bdsData = _pybdsim.Data.Load(bdsim)

    bdsinst = _pybdsim.Data.CheckItsBDSAsciiData(bdsData, True)
    bdsopt  = _GetBDSIMOptics(bdsinst)
    survey  = bdsData.model if hasattr(bdsData, "model") else None

    eTwi = _LoadSDDSColumnsToDict(elegantTwiss)
    eSig = _LoadSDDSColumnsToDict(elegantSigma)
    eCen = _LoadSDDSColumnsToDict(elegantCentroid)
    eleopt  = _GetElegantOptics(eTwi,eSig,eCen,particleType)

    figures = [PlotBeta(eleopt, bdsopt, survey=survey),
               PlotAlpha(eleopt, bdsopt, survey=survey),
               PlotDisp(eleopt, bdsopt, survey=survey),
               PlotDispP(eleopt, bdsopt, survey=survey),
               PlotSigma(eleopt, bdsopt, survey=survey),
               PlotSigmaP(eleopt, bdsopt, survey=survey),
               PlotMean(eleopt, bdsopt, survey=survey),
               PlotEmitt(eleopt, bdsopt, survey=survey),
               PlotNParticles(bdsopt, survey=survey),
               PlotEnergy(eleopt, bdsopt, survey=survey)]

    if saveAll:
        tfsname = elegantTwiss.split('/')[-1].split('.')[0]
        bdsname = repr(bdsinst)
        output_filename = "optics-report.pdf"
        if outputFileName is not None:
            output_filename = outputFileName
            if not output_filename.endswith('.pdf'):
                output_filename += ".pdf"
        else:
            output_filename = fname.replace('.root','')
            output_filename += ".pdf"
        # Should have a more descriptive name really.
        with _PdfPages(output_filename) as pdf:
            for figure in figures:
                pdf.savefig(figure)
            d = pdf.infodict()
            d['Title'] = "{} (TFS) VS {} (BDSIM) Optical Comparison".format(tfsname, bdsname)
            d['CreationDate'] = _datetime.datetime.today()

        print("Written ", output_filename)


def _GetBDSIMOptics(optics):
    """
    Takes a BDSAscii instance.
    Return a dictionary of lists matching the variable with the list of values.
    """
    optvars = {}
    for variable in optics.names:
        datum = getattr(optics, variable)()
        optvars[variable] = datum
    return optvars

def _GetElegantOptics(twiss, sigma, centroid, particleType="e-"):
    """
    Requires 3 dicts for twiss, sigma and centroid files from Elegant.
    """
    eTwissToMadx = {'ElementName':'NAME',
                    's': 'S',
                    'betax':'BETX',
                    'betay':'BETY',
                    'alphax':'ALFX',
                    'alphay':'ALFY',
                    'etax':'DX',
                    'etaxp':'DPX',
                    'etay':'DY',
                    'etayp':'DPY'}
    eSigmaToMadx = {'Sx':'SIGMAX',
                    'Sy':'SIGMAY',
                    'Sxp':'SIGMAXP',
                    'Syp':'SIGMAYP',
                    'ex':'EX',
                    'ey':'EY',
                    'Sdelta':'SIGMAP'}
    eCentroidToMadx = {'Cx':'X',
                       'Cy':'Y',
                       'Cxp':'PX',
                       'Cyp':'PY',
                       'pCentral':'P'}
    
    optvars = {}
    for kele,kmadx in eTwissToMadx.items():
        optvars[kmadx] = twiss[kele]
    for kele,kmadx in eSigmaToMadx.items():
        optvars[kmadx] = sigma[kele]
    for kele,kmadx in eCentroidToMadx.items():
        optvars[kmadx] = centroid[kele]

    massDict = {"e-" : 0.00051099891,
                "proton" : 0.938}
    massDict["e+"] = massDict["e-"]

    optvars['P'] *= massDict[particleType] # fix energy from lorentz gamma to E (yeah... P vs E..)
    return optvars

def _GetTfsOrbit(optics):
    '''
    Takes either Tfs instance.  Returns dictionary of lists.
    '''

    MADXOpticsVariables = frozenset(['S',
                                     'X',
                                     'Y',
                                     'PX',
                                     'PY'])

    optvars = {}
    for variable in MADXOpticsVariables:
        optvars[variable] = optics.GetColumn(variable)
    return optvars

# Predefined lists of tuples for making the standard plots,
# format = (optical_var_name, optical_var_error_name, legend_name)

_BETA = [("BETX", "Beta_x", "Sigma_Beta_x", r'$\beta_{x}$'),
         ("BETY", "Beta_y", "Sigma_Beta_y", r'$\beta_{y}$')]

_ALPHA = [("ALFX", "Alpha_x", "Sigma_Alpha_x", r"$\alpha_{x}$"),
          ("ALFY", "Alpha_y", "Sigma_Alpha_y", r"$\alpha_{y}$")]

_DISP = [("DX", "Disp_x", "Sigma_Disp_x", r"$D_{x}$"),
         ("DY", "Disp_y", "Sigma_Disp_y", r"$D_{y}$")]

_DISP_P = [("DPX", "Disp_xp", "Sigma_Disp_xp", r"$D_{p_{x}}$"),
           ("DPY", "Disp_yp", "Sigma_Disp_yp", r"$D_{p_{y}}$")]

_SIGMA = [("SIGMAX", "Sigma_x", "Sigma_Sigma_x", r"$\sigma_{x}$"),
          ("SIGMAY", "Sigma_y", "Sigma_Sigma_y", r"$\sigma_{y}$")]

_SIGMA_P = [("SIGMAXP", "Sigma_xp", "Sigma_Sigma_xp", r"$\sigma_{xp}$"),
            ("SIGMAYP", "Sigma_yp", "Sigma_Sigma_yp", r"$\sigma_{yp}$")]

_MEAN = [("X", "Mean_x", "Sigma_Mean_x", r"$\bar{x}$"),
         ("Y", "Mean_y", "Sigma_Mean_y", r"$\bar{y}$")]

_EMITT = [("EX", "Emitt_x", "Sigma_Emitt_x", r"$\eta_{x}$"),
          ("EY", "Emitt_y", "Sigma_Emitt_y", r"$\eta_{y}$")]

_ENERGY = [("P", "Mean_E", "Sigma_E", r"Energy")]

def _MakePlotter(plot_info_tuples, x_label, y_label, title):
    def f_out(ele, bds, survey=None, **kwargs):
        # options
        tightLayout = True
        if 'tightLayout' in kwargs:
            tightLayout = kwargs['tightLayout']

        # Get the initial N for the two sources
        first_nparticles = bds['Npart'][0]

        plot = _plt.figure(title, figsize=(9,5), **kwargs)
        colours = ('b', 'g')
        # Loop over the variables in plot_info_tuples and draw the plots.
        for a, colour in zip(plot_info_tuples, colours):
            eVar, bVar, bError, legend_name = a # unpack one tuple
            bdsS = bds['S'] # cache data
            bdsD = bds[bVar]
            bdsE = bds[bError]
            l = "{} {}; N = {:.1E}".format("", legend_name, first_nparticles),
            _plt.errorbar(bdsS, bdsD, fmt=colour+".", yerr=bdsE, label='BDSIM', capsize=3, **kwargs)
            #_plt.plot(bdsS, bdsD, colour) # line plot without label
            eS = ele['S']
            eVar = ele[eVar]
            _plt.plot(eS, eVar, colour+"--", label='Elegant')
            #_plt.plot(eS, eVar, colour+".")

        # Set axis labels and draw legend
        axes = _plt.gcf().gca()
        axes.set_ylabel(y_label)
        axes.set_xlabel(x_label)
        axes.legend(loc='best')

        if survey is not None:
            _AddMachineLatticeFromSurveyToFigure(plot, survey, tightLayout)
        else:
            _plt.tight_layout()

        _plt.show(block=False)

        return plot
    return f_out

PlotBeta   = _MakePlotter(_BETA,    "S / m", r"$\beta_{x,y}$ / m",      "Beta")
PlotAlpha  = _MakePlotter(_ALPHA,   "S / m", r"$\alpha_{x,y}$ / m",     "Alpha")
PlotDisp   = _MakePlotter(_DISP,    "S / m", r"$D_{x,y} / m$",          "Dispersion")
PlotDispP  = _MakePlotter(_DISP_P,  "S / m", r"$D_{p_{x},p_{y}}$ / m",  "Momentum_Dispersion")
PlotSigma  = _MakePlotter(_SIGMA,   "S / m", r"$\sigma_{x,y}$ / m",     "Sigma")
PlotSigmaP = _MakePlotter(_SIGMA_P, "S / m", r"$\sigma_{xp,yp}$ / rad", "SigmaP")
PlotMean   = _MakePlotter(_MEAN,    "S / m", r"$\bar{x}, \bar{y}$ / m", "Mean")
PlotEmitt  = _MakePlotter(_EMITT,   "S / m", r"$\eta_{x}, \eta_{y}$ / rad m", "Emittance")
PlotEnergy = _MakePlotter(_ENERGY,  "S / m", r"Total Energy / GeV",     "Energy")

def PlotNParticles(bdsopt, survey=None, **kwargs):
    # options
    tightLayout = kwargs['tightLayout'] if 'tightLayout' in kwargs else True
    
    npartPlot = _plt.figure('NParticles', figsize=(9,5), **kwargs)

    _plt.plot(bdsopt['S'],bdsopt['Npart'], 'k-', label='BDSIM N Particles')
    _plt.plot(bdsopt['S'],bdsopt['Npart'], 'k.')
    axes = _plt.gcf().gca()
    axes.set_ylabel(r'N Particles')
    axes.set_xlabel('S / m')
    axes.legend(loc='best')

    if survey is not None:
        _AddMachineLatticeFromSurveyToFigure(npartPlot, survey, tightLayout)
    else:
        _plt.tight_layout()

    _plt.show(block=False)
    return npartPlot

def _CheckFileExistsList(*fns):
    for fn in fns:
        _CheckFileExists(fn)

def _CheckFileExists(fn):
    if isinstance(fn, str):
        if not _isfile(fn):
            raise IOError('File "'+fn+'" not found')



from .. import Data as _Data
from .. import Convert as _Convert
from .. import Plot as _Plot
import pymadx as _pymadx
import os as _os
import matplotlib.pyplot as _plt
import matplotlib.backends.backend_pdf as _pdf
import matplotlib.patches as _mpatches
import numpy as _np
import subprocess as _sub
import threading as _thread
import time as _time

from scipy.stats import binned_statistic as _bin
import scipy.constants as _scpconsts

class LatticeTest:
    def __init__(self,filepath, nparticles = 1000, verbose=False):        
        """
        Takes a .madx file containing description of a lattice and generates
        BDSIM and MadX PTC jobs from it as well as MadX optical functions 
        propagation plots.

        nparticles: specifies the number of particles to be used for BDSIM/PTC runs. Default = 1000
        verbose   : prints additional information
        """
        cwd        = _os.getcwd()
        self.existingfiles = _os.listdir(cwd)
        path       = filepath.split("/")
        filename   = path[-1]                         #last element is the filename
        
        if(path[0]=="/"):                             #when absolute filepath is given
            folderpath = path[:-1]
        elif(path[0]=="." and path[1]=="/"):          #when ./ is used to specify current folder
            path=path[1:]
            folderpath = cwd+"/".join(path[:-1])      #when relative filepath is given
        else:
            folderpath = cwd+"/"+"/".join(path[:-1])  #when relative filepath is given

        print("Filename: ",filename)
        print("Folderpath: ",folderpath)
        
        if filename[-5:] == ".madx":                
            self.filename    = filename[:-5]
            self.tfsfilename = str.lower(self.filename)
            self.filepath    = filepath
            self.folderpath  = folderpath
            self.ptcinrays   = self.filename+"_inrays.madx"
            self.ptcfilename = "ptc_"+self.filename+".madx"
            self.nparticles  = nparticles
            self.verbose     = verbose
            self.flipmagnets = False
            self.figureNr    = 1729
        else:
            raise IOError('Invalid file format, MadX file (.madx) required')

    def CleanRunCompare(self):
        self.Clean()
        self.Run()
        self.Compare(addPrimaries=False)

    def Clean(self):
        """
        Delete all files produced during the testing, including
        .log .dat .tfs. .ps ptc* .txt .root .gmad inrays .png .pdf
        """
        _os.chdir(self.folderpath)

        
        #PTC files
        ptcMain         = "ptc_"+self.filename+".madx"
        ptcComponents   = "ptc_"+self.filename+"_components.madx"
        ptcJob          = "ptc_"+self.filename+"_ptcjob.madx"
        ptcSequence     = "ptc_"+self.filename+"_sequence.madx"
        ptcBeam         = "ptc_"+self.filename+"_beam.madx"
        
        #gmad files
        gmadMain        = self.filename+".gmad"
        gmadComponents  = self.filename+"_components.gmad"
        gmadOptions     = self.filename+"_options.gmad"
        gmadSequence    = self.filename+"_sequence.gmad"
        gmadBeam        = self.filename+"_beam.gmad"
        
        #bdsim output
        event           = self.filename+".root"
        optics          = self.filename+"_optics.root"
        config          = self.filename+"_analConfig.txt"

        #ptc/madx output
        tfs             = self.filename.lower() + ".tfs"
        trackone        = "trackone"
        maxwell         = "Maxwellian*"
        printFile       = "print.dat"
        ptcOpt          = "ptc_"+self.filename+"_opticalfns.dat"
        
        #logfiles
        bdsimLog        = "bdsim.log"
        madxLog         = "madx.log"
        ptcLog          = "ptc_madx.log"
        
        #plots
        psFile          = "madx.ps"
        pdfFile         = self.filename+"_plots.pdf"
        
        #other
        inrays          = self.filename+"_inrays.madx"
        stdevFile       = self.filename+"_stdev.txt"
        
        try:
            _os.remove(ptcMain)
            _os.remove(ptcComponents)
            _os.remove(ptcJob)
            _os.remove(ptcSequence)
            _os.remove(ptcBeam)
            _os.remove(gmadMain)
            _os.remove(gmadComponents)
            _os.remove(gmadOptions)
            _os.remove(gmadSequence)
            _os.remove(gmadBeam)
            _os.remove(event)
            _os.remove(optics)
            _os.remove(config)
            _os.remove(tfs)
            _os.remove(trackone)
            _os.system("rm -rf "+maxwell)
            _os.remove(printFile)
            _os.remove(ptcOpt)
            _os.remove(bdsimLog)
            _os.remove(madxLog)
            _os.remove(ptcLog)
            _os.remove(psFile)
            _os.remove(pdfFile)
            _os.remove(inrays)
            _os.remove(stdevFile)

        except OSError:
            # File probably doesn't exist
            pass

#        _os.system("rm -rf *.log")
#        _os.system("rm -rf *.dat")
#        _os.system("rm -rf *.tfs")
#        _os.system("rm -rf *.ps")
#        _os.system("rm -rf ptc*")
#        _os.system("rm -rf *.txt")
#        _os.system("rm -rf *.root")
#        _os.system("rm -rf *.gmad")
#        _os.system("rm -rf *_inrays.madx")
#        _os.system("rm -rf *.png")
#        _os.system("rm -rf *.pdf")
#        _os.system("rm trackone")

        # clean and close figures (10 figures in total)
        for i in range(11):
            _plt.close(self.figureNr+i)
            

    def CleanOutput(self):
        """
        Delete output files files produced during the testing, while
        preserving lattice definitions. Useful for beam parameter testing.
        Deleted files include .dat .txt .root inrays .png .pdf
        """
        _os.chdir(self.folderpath)      
        _os.system("rm -rf *.dat")
        _os.system("rm -rf *.root")
        _os.system("rm -rf *.txt")
        _os.system("rm -rf *.png")
        _os.system("rm -rf *.pdf")
        _os.system("rm -rf *.log")
        _os.system("rm -rf *.ps")
        _os.system("rm -rf *_inrays.madx")
        _os.system("rm -rf *.gmad")
        _os.system("rm -rf ptc_*")
        _os.system("rm -rf *.tfs")
        _os.system("rm trackone")

        # clean and close figures (10 figures in total)
        for i in range(11):
            _plt.close(self.figureNr+i)

    def CleanNewFiles(self):
        """
        Delete new files produced by the current running instance of the class and leaves 
        all files orginially contained in the folder intact.
        """
        _os.chdir(self.folderpath)
        newfiles = _os.listdir(_os.getcwd())
        
        for oldfile in self.existingfiles:
            for newfile in newfiles:
                if newfile == oldfile:
                    newfiles.remove(newfile)
        
        for file in newfiles:
            _os.remove(file)

        # clean and close figures (10 figures in total)
        for i in range(11):
            _plt.close(self.figureNr+i)

    def RunParams(self, bdsim='bdsim', madx='madx'):
        """
        Allows a bdsim and ptc run with already defined lattices. Only updates the beam distribution and
        produces the new output.
        """
        print('Test> Running new parameters: ', self.filename)

        _os.chdir(self.folderpath)
        _os.system(bdsim+" --file="+self.filename+".gmad --ngenerate="+str(self.nparticles)+" --batch --output=rootevent --outfile="+self.filename+"> bdsim.log")
        _Convert.BdsimPrimaries2Ptc(''+self.filename+'.root', self.ptcinrays)
        _os.system(madx+" < "+self.ptcfilename+" > ptc_madx.log")


    def Run(self,
            bdsim='bdsim',
            madx='madx',
            integratorSet=None,
            optionsdict=None):
        optionsdict = {} if optionsdict is None else optionsdict
        print('Test> Lattice: ', self.filename) 
        print('Test> Destination filepath: ', self.filepath)

        _os.chdir(self.folderpath)

        _os.system(madx + " < "+self.filename+".madx > madx.log")

        """
        a = pybdsim.Convert.MadxTfs2Gmad(''+self.tfsfilename+'.tfs','dump',beam=False)
        b = pybdsim.Beam.Beam('proton',10.0,'ptc')    #beam parameters need to be set manually
        b.SetDistribFileName('INRAYS.madx')           #This is for testing BDSIM 'ptc' beam distribution
        a.AddBeam(b)
        a.Write(self.filename)
        
        """
        #Load Tfs file to check particle type and flip BDSIM magnet polarities as needed
        tfs  = _pymadx.Data.Tfs(self.tfsfilename+'.tfs')
        particle = tfs.header['PARTICLE']
        if particle == 'ELECTRON' :
            self.flipmagnets = True

        beamparamsdict = {'offsetSampleMean': 1}
        if integratorSet is not None:
            optionsdict['integratorSet'] = '"'+integratorSet+'"'
        _Convert.MadxTfs2Gmad(self.tfsfilename+'.tfs',
                              self.filename,
                              flipmagnets=self.flipmagnets,
                              ignorezerolengthitems=False,
                              verbose=self.verbose,
                              optionsdict=optionsdict,
                              beamparamsdict=beamparamsdict)
        
        _pymadx.Convert.TfsToPtc(''+self.tfsfilename+'.tfs', self.ptcfilename, self.ptcinrays, ignorezerolengthitems=False, ptctrackaperture=[1.0, 1.0, 1.0, 1.0, 1.0, 1.0])

        # run process through subprocess module. Safer than running BDSIM through os.system which can cause problems.
        process = _sub.Popen([bdsim,
                        "--file=" + self.filename + ".gmad",
                        "--outfile="+self.filename,
                        "--ngenerate="+str(self.nparticles),
                        "--batch",
                        "--seed=1993"],
                        stdout=open('bdsim.log', 'a'),
                        stderr=open('bdsim.log', 'a'))

        # Method of communicating with BDSIM process. Start and apply the timeout via joining
        processThread = _thread.Thread(target=process.communicate)
        processThread.start()
        processThread.join()

        #_os.system(bdsim+" --file="+self.filename+".gmad --ngenerate="+str(self.nparticles)+" --batch --seed=1993 --output=rootevent --outfile="+self.filename+"> bdsim.log")

        _Convert.BdsimPrimaries2Ptc(''+self.filename+'.root', self.ptcinrays)

        _os.system(madx+" < "+self.ptcfilename+" > ptc_madx.log")


    def Compare(self, plot='all', addPrimaries=False, showPlots=False, showResiduals=True, rebdsim="rebdsim", noPlots=False):
        """
        Performs analysis and comparison of BDSIM, MADX and MADX-PTC output. 
       
        addPrimaries - True adds BDSIM primaries to histos. Default is False
        plot         - | beta | emittance | alpha |  - specify optical function to plot
        showPlots    - True diplays the plots to screen
        """

        _os.chdir(self.folderpath)

        #Writes the text file for Rebdsim
        with open(''+self.filename+'_analConfig.txt', 'w') as outfile:
            outfile.writelines("{:<40s}".format('Debug')+'\t 0\n')
            outfile.writelines("{:<40s}".format('InputFilePath')+'\t ./'+self.filename+'.root \n')
            outfile.writelines("{:<40s}".format('OutputFileName')+'\t ./'+self.filename+'_optics.root \n')
            outfile.writelines("{:<40s}".format('CalculateOpticalFunctions')+'\t 1 \n')
            outfile.writelines("{:<40s}".format('CalculateOpticalFunctionsFileName')+'\t ./'+self.filename+'_optics.dat \n')
            outfile.writelines("{:<40s}".format('emittanceOnTheFly')+'\t 1 \n')

        #Calculates optical functions and produces .root and .dat files for analysis
        #_os.system(rebdsim+" "+self.filename+"_analConfig.txt")
        process = _sub.Popen([rebdsim, self.filename+"_analConfig.txt"])

        # Method of communicating with BDSIM process. Start and apply the timeout via joining
        processThread = _thread.Thread(target=process.communicate)
        processThread.start()
        processThread.join()

        #Load data
        f = _Data.Load(self.filename+".root")
        last_name = f.GetSamplerNames()[-1]
        sampler = _Data.SamplerData(f, last_name)
        primary = _Data.SamplerData(f)

        # Get the sampler particle coordinates.
        Bx      = sampler.data['x']
        Bxp     = sampler.data['xp']
        By      = sampler.data['y']
        Byp     = sampler.data['yp']
        Btof    = sampler.data['T']
        BE      = sampler.data['energy']
        meanE   = _np.mean(BE)
        meantof = _np.mean(Btof)

        #Get particle pdg number
        priPid = primary.data['partID']
        pid    =  _np.int(_np.mean(priPid))  #cast to int to match pdg id

        #Particle mass needed for calculating momentum, in turn needed for dE.
        #All in GeV, taken from the PDG table http://pdg.lbl.gov/2017/reviews/rpp2017-rev-phys-constants.pdf
        if pid == 2212:                                     #proton
            mass = 0.9382720813
        elif (pid == 11) or (pid == -11):                   #electron / positron
            mass = 0.5109989461e-3
        elif (pid == 13) or (pid == -13):                   #mu- / mu+
            mass = 0.1056583745

        else:
            raise ValueError("Mass for particle with PDG ID {} not available, stop".format(pid)
            )

        beta = _np.sqrt(1 - (mass/meanE)**2)

        #get canonical coordinate t (time_offset_from_reference_part*speed_of_light) from
        #the recorded time of flight in bdsim
        Bt         = beta*(Btof - _np.full_like(Btof, meantof))*1.e-9*_scpconsts.c

        P = _np.sqrt(_np.square(BE) - _np.full_like(BE, mass**2))
        p0 = _np.sqrt(_np.mean(BE)**2 - 0.93827208148**2)
        BPT = (P - _np.full_like(P, p0))/p0

        self.bdsimoutput = {'x':Bx,'y':By,'xp':Bxp,'yp':Byp, 'pt':BPT, 't':Bt}

        madxout = _pymadx.Data.Tfs("trackone")
        madxend = madxout.GetSegment(madxout.nsegments) #(1) #get the last 'segment' / sampler
        Mx = madxend.GetColumn('X')
        My = madxend.GetColumn('Y')
        Mxp = madxend.GetColumn('PX')
        Myp = madxend.GetColumn('PY')

        MPT = madxend.GetColumn('PT')
        MT  = madxend.GetColumn('T')
        self.ptcoutput = {'x':Mx,'y':My,'xp':Mxp,'yp':Myp, 'pt':MPT, 't':MT}

        #Check all particles make it through
        if(len(Bx)!=len(Mx)):       

            print("bdsimMadx.Compare>>Error: Unequal number of output particles, please check parameters and try again")
            print("Input particles: ",self.nparticles," BDS_out particles: ", len(Bx), " PTC_out particles: ", len(Mx))
            return
                
        if showResiduals:
            # residuals
            fresx  = Mx - Bx
            fresy  = My - By
            fresxp = Mxp - Bxp
            fresyp = Myp - Byp
            self.residuals = {'x':fresx,'y':fresy,'xp':fresxp,'yp':fresyp}
    
            rmsx = "%.6E" % _np.sqrt(_np.average(fresx*fresx))
            rmsy = "%.6E" % _np.sqrt(_np.average(fresy*fresy))
            rmsxp = "%.6E" % _np.sqrt(_np.average(fresxp*fresxp))
            rmsyp = "%.6E" % _np.sqrt(_np.average(fresyp*fresyp))

            print("Rms residuals:")
            print(" x(m):    ", rmsx,  " y(m):    ", rmsy)
            print(" xp(rad): ", rmsxp, " yp(rad): ", rmsyp)
            
            #standard deviation
            stdMx  = _np.std(Mx)
            stdMy  = _np.std(My)
            stdMxp = _np.std(Mxp)
            stdMyp = _np.std(Myp)

            stdBx  = _np.std(Bx)
            stdBy  = _np.std(By)
            stdBxp = _np.std(Bxp)
            stdByp = _np.std(Byp)

            #standard devation fractional errors
            frestdx  = _np.nan_to_num(stdMx - stdBx)
            frestdy  = _np.nan_to_num(stdMy - stdBy)
            frestdx  = _np.nan_to_num(frestdx / stdMx) #protect against nans for 0 diff
            frestdy  = _np.nan_to_num(frestdy / stdMy)
            frestdxp = _np.nan_to_num(stdMxp - stdBxp)
            frestdyp = _np.nan_to_num(stdMyp - stdByp)
            frestdxp = _np.nan_to_num(frestdxp / stdMxp)
            frestdyp = _np.nan_to_num(frestdyp / stdMyp)

            # write standard deviations to file
            with open(''+self.filename+'_stdev.txt', 'w') as stdout:
                timestamp = _time.strftime("%Y/%m/%d-%H:%M:%S")
                t = timestamp+' '+self.filename+' Standard Deviations (particles = '+str(self.nparticles)+'): \n'
                h = 'BDSIM_X \t MX-PTC_X \t BDSIM_Y \t MX-PTC_Y \t BDSIM_XP'
                h+= ' \t MX-PTC_XP \t BDSIM_YP \t MX-PTC_YP \t FRCERR_X \t FRCERR_Y \t FRCERR_XP \t FRCERR_YP \n'
                s  = "{0:.4e}".format(stdBx)
                s += '\t' +  "{0:.4e}".format(stdMx)
                s += '\t' +  "{0:.4e}".format(stdBy)              
                s += '\t' +  "{0:.4e}".format(stdMy)
                s += '\t' +  "{0:.4e}".format(stdBxp)
                s += '\t' +  "{0:.4e}".format(stdMxp)
                s += '\t' +  "{0:.4e}".format(stdByp)
                s += '\t' +  "{0:.4e}".format(stdMyp)
                s += '\t' +  "{0:.4e}".format(frestdx)
                s += '\t' +  "{0:.4e}".format(frestdy)
                s += '\t' +  "{0:.4e}".format(frestdxp)
                s += '\t' +  "{0:.4e}".format(frestdyp) + '\n'
                stdout.writelines(t)
                stdout.writelines(h)
                stdout.writelines(s)            
        
        #Loading output and processing optical functions
        madx = _pymadx.Data.Tfs(''+self.tfsfilename+'.tfs')

        #Writes the text file for Rebdsim
        with open(''+self.filename+'_analConfig.txt', 'w') as outfile:
            outfile.writelines("{:<40s}".format('Debug')+'\t 0\n')
            outfile.writelines("{:<40s}".format('InputFilePath')+'\t ./'+self.filename+'.root \n')
            outfile.writelines("{:<40s}".format('OutputFileName')+'\t ./'+self.filename+'_optics.root \n')
            outfile.writelines("{:<40s}".format('CalculateOpticalFunctions')+'\t 1 \n')
            outfile.writelines("{:<40s}".format('CalculateOpticalFunctionsFileName')+'\t ./'+self.filename+'_optics.dat \n')
            outfile.writelines("{:<40s}".format('emittanceOnTheFly')+'\t 0 \n')

        #Calculates optical functions and produces .root and .dat files for analysis 
        #_os.system(rebdsim+" "+self.filename+"_analConfig.txt")
        process = _sub.Popen([rebdsim, self.filename+"_analConfig.txt"])

        # Method of communicating with BDSIM process. Start and apply the timeout via joining
        processThread = _thread.Thread(target=process.communicate)
        processThread.start()
        processThread.join()

        if noPlots:
            return

        bdata  = _Data.Load(self.filename+'_optics.root').optics

        #ptcfile  = 'ptc_'+self.filename+'_opticalfns.dat'
        #print "ptcCalculateOpticalFunctions> processing... " , ptcfile 
        #ptc      = _pymadx.PtcAnalysis.PtcAnalysis(ptcOutput="trackone") 
        #ptc.CalculateOpticalFunctions(ptcfile)
        #ptcdata  = pybdsim.Data.Load(ptcfile)

        #Get the S coordinate from all outputs
        M_s       = madx.GetColumn('S')
        B_s       = bdata.S()
        #PTC_s     = ptcdata.S()

        M_emittx  = madx.header['EX'] #Get emittance from tfs file header
        M_emitty  = madx.header['EY'] #To be used in emittance and sigma plots

        plotopts = [] #keep track of plots
        plotNr   = 0

        if  (plot == "all"):
            plotopts.append("beta")
            plotopts.append("dispersion_xy")
            plotopts.append("dispersion_xpyp")
            plotopts.append("alpha")
            plotopts.append("emittance")
            plotopts.append("sigma_xy")
            plotopts.append("sigma_xpyp")
        else:
            plotopts.append(plot)

        

        #optfn denotes the selected optical function to plot
        in_Tfs = True       #some of the calculated optical functions are not present in the tfs file (e.g emittance,sigmas)
                            #and hence plots and residuals between BDSIM and MADX cannot be obtained
        for opt in plotopts:
            print("bdsimMadx... plotting ", opt)
            if (opt=='beta'):
                fn_name      = r'\beta' #this is a raw string for Latex labels and titles
                fn_rname     = 'beta'   #this is reduced name for filename of saved figure
                fn_subsc     = ['_{x}','_{y}']
                fn_units     = '(m)'
                plotNr      += 1
                
                M_optfn_x    = madx.GetColumn('BETX') 
                M_optfn_y    = madx.GetColumn('BETY')
                B_optfn_x    = bdata.Beta_x()
                B_optfn_y    = bdata.Beta_y()
                B_opterr_x   = bdata.Sigma_Beta_x()
                B_opterr_y   = bdata.Sigma_Beta_y()
                #PTC_optfn_x  = ptcdata.Beta_x()
                #PTC_optfn_y  = ptcdata.Beta_y()
                #PTC_opterr_x = ptcdata.Sigma_beta_x()
                #PTC_opterr_y = ptcdata.Sigma_beta_y()
                
                #print 'LenMopt ',len(M_optfn_x),' LenMs ',len(M_s),' LenBs ',len(B_s),' LenBopt ',len(B_optfn_x)
                
            elif (opt=='alpha'):
                fn_name      = r'\alpha' #this is a raw string for Latex labels and titles
                fn_rname     = 'alpha'   #this is reduced name for filename of saved figure
                fn_subsc     = ['_{x}','_{y}']
                fn_units     = ''
                plotNr      += 1
                
                M_optfn_x    = madx.GetColumn('ALFX') 
                M_optfn_y    = madx.GetColumn('ALFY')
                B_optfn_x    = bdata.Alpha_x()
                B_optfn_y    = bdata.Alpha_y()
                B_opterr_x   = bdata.Sigma_Alpha_x()
                B_opterr_y   = bdata.Sigma_Alpha_y()
                #PTC_optfn_x  = ptcdata.Alph_x()
                #PTC_optfn_y  = ptcdata.Alph_y()
                #PTC_opterr_x = ptcdata.Sigma_alph_x()
                #PTC_opterr_y = ptcdata.Sigma_alph_y()

            elif (opt=='sigma_xy'):
                fn_name      = r'\sigma'  
                fn_rname     = 'sigma_xy' 
                fn_subsc     = ['_{x}','_{y}']
                fn_units     = '(m)'
                plotNr      += 1
                
                in_Tfs       = True
                M_optfn_x    = madx.GetColumn('SIGMAX')
                M_optfn_y    = madx.GetColumn('SIGMAY')
                B_optfn_x    = bdata.Sigma_x()
                B_optfn_y    = bdata.Sigma_y()
                B_opterr_x   = bdata.Sigma_Sigma_x()
                B_opterr_y   = bdata.Sigma_Sigma_y()
                #PTC_optfn_x  = ptcdata.Sigma_x()
                #PTC_optfn_y  = ptcdata.Sigma_y()
                #PTC_opterr_x = 0.000001 #not implemented in PTC yet
                #PTC_opterr_y = 0.000001
                
            elif (opt=='sigma_xpyp'):
                fn_name      = r'\sigma'
                fn_rname     = 'sigma_xpyp'
                fn_subsc     = ['_{xp}','_{yp}']
                fn_units     = '(rad)'
                plotNr      += 1

                in_Tfs       = True
                M_optfn_x    = madx.GetColumn('SIGMAXP')
                M_optfn_y    = madx.GetColumn('SIGMAYP')
                B_optfn_x    = bdata.Sigma_xp()
                B_optfn_y    = bdata.Sigma_yp()
                B_opterr_x   = bdata.Sigma_Sigma_xp()
                B_opterr_y   = bdata.Sigma_Sigma_yp()
                #PTC_optfn_x  = ptcdata.Sigma_xp()
                #PTC_optfn_y  = ptcdata.Sigma_yp()
                #PTC_opterr_x = 0.000001 #not implemented in PTC yet
                #PTC_opterr_y = 0.000001       

            elif (opt=='emittance'):
                fn_name      = r'\epsilon' #this is a raw string for Latex labels and titles
                fn_rname     = 'emittance'   #this is reduced name for filename of saved figure
                fn_subsc     = ['_{x}','_{y}']
                fn_units     = '(m)'
                plotNr      += 1
                
                in_Tfs       = True
                M_optfn_x    = _np.empty(len(M_s)); #Emittance in the TFS file is a constant defined in the header
                M_optfn_x.fill(M_emittx)       
                M_optfn_y    = _np.empty(len(M_s));
                M_optfn_y.fill(M_emitty)
                B_optfn_x    = bdata.Emitt_x()
                B_optfn_y    = bdata.Emitt_y()
                B_opterr_x   = bdata.Sigma_Emitt_x()
                B_opterr_y   = bdata.Sigma_Emitt_y()
                #PTC_optfn_x  = ptcdata.Emitt_x()
                #PTC_optfn_y  = ptcdata.Emitt_y()
                #PTC_opterr_x = ptcdata.Sigma_emitt_x()
                #PTC_opterr_y = ptcdata.Sigma_emitt_y()

            elif (opt=='dispersion_xy'):
                fn_name      = r'D'              #this is a raw string for Latex labels and titles
                fn_rname     = 'dispersion_xy'   #this is reduced name for filename of saved figure
                fn_subsc     = ['_{x}','_{y}']
                fn_units     = '(m)'
                plotNr      += 1

                in_Tfs       = True
                M_optfn_x    = madx.GetColumn('DXBETA')
                M_optfn_y    = madx.GetColumn('DYBETA')
                B_optfn_x    = bdata.Disp_x()
                B_optfn_y    = bdata.Disp_y()
                #PTC_optfn_x  = ptcdata.Disp_x()
                #PTC_optfn_y  = ptcdata.Disp_y()
                #PTC_opterr_x = 0 #error calculations for dispersion not implemented yet
                #PTC_opterr_y = 0
                B_opterr_x   = bdata.Sigma_Disp_x()
                B_opterr_y   = bdata.Sigma_Disp_y()

            elif (opt=='dispersion_xpyp'):
                fn_name      = r'D' #this is a raw string for Latex labels and titles
                fn_rname     = 'dispersion_xpyp'   #this is reduced name for filename of saved figure
                fn_subsc     = ['_{xp}','_{yp}']
                fn_units     = '(rad)'
                plotNr      += 1

                print("Warning: Disp_xpyp not present in MADX tfs file, plotting only MADX-PTC and BDSIM results ")
                in_Tfs       = True
                M_optfn_x    = madx.GetColumn('DPXBETA')
                M_optfn_y    = madx.GetColumn('DPYBETA')
                B_optfn_x    = bdata.Disp_xp()
                B_optfn_y    = bdata.Disp_yp()
                #PTC_optfn_x  = ptcdata.Disp_xp()
                #PTC_optfn_y  = ptcdata.Disp_yp()
                #PTC_opterr_x = 0 #error calculations for dispersion not implemented yet
                #PTC_opterr_y = 0
                B_opterr_x   = bdata.Sigma_Disp_xp()
                B_opterr_y   = bdata.Sigma_Disp_yp()
            else:
                print("Error: Unrecognised plotting option:", plot)
                return
            
            #   if(in_Tfs):  
            #       M_s    = M_s[:len(B_s)]
            #       M_optfn_x = M_optfn_x[:len(B_optfn_x)]    #Madx arrays need to be sliced because they contain
            #       M_optfn_y = M_optfn_y[:len(B_optfn_y)]    #one too many columns. No information is lost in the slicing
            #as the last Madx segment is default end of the line info and is
            #degenerate with the last element segment
            #  PTC_s    = PTC_s[:len(B_s)]
            #  PTC_optfn_x = PTC_optfn_x[:len(B_optfn_x)]
            #  PTC_optfn_y = PTC_optfn_y[:len(B_optfn_y)]
        
            _plt.figure(self.figureNr+plotNr-2, figsize=(13, 8), dpi=80, facecolor='w', edgecolor='k')
            _plt.clf()
            _plt.plot([], [], color='w', linewidth=0.001,alpha=0.1,label=r'Npart= '+str(self.nparticles))
            if(in_Tfs):
                _plt.plot(M_s,M_optfn_x,'.',color='r',linestyle='dashed',linewidth=2.0,label=r'$'+fn_name+fn_subsc[0]+r'$MDX')
                _plt.plot(M_s,M_optfn_y,'.',color='b',linestyle='dashed',linewidth=2.0,label=r'$'+fn_name+fn_subsc[1]+r'$MDX')
            #_plt.fill_between(PTC_s, PTC_optfn_x-PTC_opterr_x, PTC_optfn_x+PTC_opterr_x,alpha=0.3, facecolor='r',linewidth=0.0)
            #_plt.fill_between(PTC_s, PTC_optfn_y-PTC_opterr_y, PTC_optfn_y+PTC_opterr_y,alpha=0.3, facecolor='b',linewidth=0.0)
            _plt.errorbar(B_s,B_optfn_x, yerr=B_opterr_x,fmt='o',color='r',label=r'$'+fn_name+fn_subsc[0]+r'$BDS')
            _plt.errorbar(B_s,B_optfn_y, yerr=B_opterr_y,fmt='o',color='b',label=r'$'+fn_name+fn_subsc[1]+r'$BDS')
            #_plt.plot([], [], color='r', linewidth=10,alpha=0.3,label=r'$'+fn_name+fn_subsc[0]+r'$PTC')
            #_plt.plot([], [], color='b', linewidth=10,alpha=0.3,label=r'$'+fn_name+fn_subsc[1]+r'$PTC')
            _plt.xlabel(r'$S (m)$')
            _plt.ylabel(r'$'+fn_name+fn_units+r'$')
            _plt.legend(numpoints=1,loc=10,fancybox=True, framealpha=1.0,prop={'size':15})
            _plt.grid(True)
            _pymadx.Plot.AddMachineLatticeToFigure(_plt.gcf(),madx)
            #_plt.subplots_adjust(left=0.1,right=0.9,top=0.96, bottom=0.15, wspace=0.15, hspace=0.2)

        if showResiduals:
            #optical function residuals
            """
            if(in_Tfs):
                res_optfn_x = M_optfn_x-B_optfn_x        
                res_optfn_y = M_optfn_y-B_optfn_y
            
                _plt.figure(self.figureNr+1, figsize=(11, 8), dpi=80, facecolor='w', edgecolor='k')
                _plt.clf()
                _plt.plot(M_s,res_optfn_x,'*',color='r',linestyle='solid',label=r'$\beta_{x}$Res')
                _plt.plot(M_s,res_optfn_y,'*',color='b',linestyle='solid',label=r'$\beta_{y}$Res')
                _plt.title(self.filename+r' Plot of $'+fn_name+r'_{x,y}$ Residuals vs $S$')
                _plt.xlabel(r'$S (m)$')
                _plt.ylabel(r'$'+fn_name+r'_{x,y}$ Residuals(m)')
                _plt.legend(numpoints=1,loc=7,prop={'size':15})
                _plt.grid(True)
            """
                   
            # 2d plots
            arrow_width_scale = 1.e-3  #Factor used to multiply minimum residual between BDSIM and PTC data in order to
                                      #heuristicaly obtain a width for the quiver plot arrows connecting the two data sets
                                      #Very crude, fix in the future

            #Scatter plots with quiver plots underneath. Useful visualising distributions of low number of particles, ignore for many particles
            if(self.nparticles<600):
                f = _plt.figure(self.figureNr+2, figsize=(13, 8), dpi=80, facecolor='w', edgecolor='k')
                f.suptitle(self.filename)
                _plt.clf()
                #X vs Y
                ax1 = f.add_subplot(221)
                arrow_width = abs(_np.min(fresy))*arrow_width_scale
                _plt.quiver(Mx,My,-fresx,-fresy,angles='xy',scale_units='xy',scale=1,color='r',units='x',width=arrow_width,headwidth=3)
                #_plt.plot(Mx,My,'b.',label='PTC')
                #_plt.plot(Bx,By,'g.',label='BDSIM')
                _plt.scatter(Mx,My,label='PTC',s=10,facecolors='red', edgecolors="red")
                _plt.scatter(Bx,By,label='BDSIM',s=60,facecolors='none',edgecolors="blue")
                if addPrimaries:
                    _plt.plot(Bx0,By0,'r.',label='BDSIM prim')
                _plt.legend(prop={'size':10})
                _plt.xlabel(r"x (m)")
                _plt.ylabel(r"y (m)")
                startx, endx = ax1.get_xlim()
                starty, endy = ax1.get_ylim()
                ax1.xaxis.set_ticks([startx,0,endx])
                ax1.yaxis.set_ticks([starty,0,endy])
                locs,labels = _plt.xticks()
                _plt.xticks(locs, ["%2.1e" % x for x in locs])
                locs,labels = _plt.yticks()
                _plt.yticks(locs, ["%2.1e" % x for x in locs])
                ax1.tick_params(axis='both', which='major', pad=7)

                #XP vs YP
                ax2 = f.add_subplot(222)
                arrow_width = abs(_np.min(fresyp))*arrow_width_scale
                _plt.quiver(Mxp,Myp,-fresxp,-fresyp,angles='xy',scale_units='xy',scale=1,color='r',units='x',width=arrow_width,headwidth=3)
                #_plt.plot(Mxp,Myp,'b.',label='PTC')
                #_plt.plot(Bxp,Byp,'g.',label='BDSIM')
                _plt.scatter(Mxp,Myp,label='PTC',s=10,facecolors='red', edgecolors="red")
                _plt.scatter(Bxp,Byp,label='BDSIM',s=60,facecolors='none',edgecolors="blue")
                if addPrimaries:
                    _plt.plot(Bxp0,Byp0,'r.',label='BDSIM prim')
                _plt.legend(prop={'size':10})
                _plt.xlabel(r"xp (rad)")
                _plt.ylabel(r"yp (rad)")
                startx, endx = ax2.get_xlim()
                starty, endy = ax2.get_ylim()
                ax2.xaxis.set_ticks([startx,0,endx])
                ax2.yaxis.set_ticks([starty,0,endy])
                locs,labels = _plt.xticks()
                _plt.xticks(locs, ["%2.1e" % x for x in locs])
                locs,labels = _plt.yticks()
                _plt.yticks(locs, ["%2.1e" % x for x in locs])
                ax2.tick_params(axis='both', which='major', pad=7)
                        
                #X vs XP
                arrow_width = abs(_np.min(fresxp))*arrow_width_scale
                ax3 = f.add_subplot(223)
                _plt.quiver(Mx,Mxp,-fresx,-fresxp,angles='xy',scale_units='xy',scale=1,color='r',units='x',width=arrow_width,headwidth=3)
                #_plt.plot(Mx,Mxp,'b.',label='PTC')
                #_plt.plot(Bx,Bxp,'g.',label='BDSIM')
                _plt.scatter(Mx,Mxp,label='PTC',s=10,facecolors='red', edgecolors="red")
                _plt.scatter(Bx,Bxp,label='BDSIM',s=60,facecolors='none',edgecolors="blue")
                if addPrimaries:
                    _plt.plot(Bx0,Bxp0,'r.',label='BDSIM prim')
                _plt.legend(prop={'size':10})
                _plt.xlabel(r"x (m)")
                _plt.ylabel(r"xp (rad)")
                startx, endx = ax3.get_xlim()
                starty, endy = ax3.get_ylim()
                ax3.xaxis.set_ticks([startx,0,endx])
                ax3.yaxis.set_ticks([starty,0,endy])
                locs,labels = _plt.xticks()
                _plt.xticks(locs, ["%2.1e" % x for x in locs])
                locs,labels = _plt.yticks()
                _plt.yticks(locs, ["%2.1e" % x for x in locs])
                ax3.tick_params(axis='both', which='major', pad=7)
                    
                #Y vs YP
                arrow_width = abs(_np.min(fresyp))*arrow_width_scale
                ax4 = f.add_subplot(224)
                _plt.quiver(My,Myp,-fresy,-fresyp,angles='xy',scale_units='xy',scale=1,color='r',units='x',width=arrow_width,headwidth=3)
                #_plt.plot(My,Myp,'b.',label='PTC')
                #_plt.plot(By,Byp,'g.',label='BDSIM')
                _plt.scatter(My,Myp,label='PTC',s=10,facecolors='red', edgecolors="red")
                _plt.scatter(By,Byp,label='BDSIM',s=60,facecolors='none',edgecolors="blue")
                if addPrimaries:
                    _plt.plot(By0,Byp,'r.',label='BDSIM prim')
                _plt.legend(prop={'size':10})
                _plt.xlabel(r"y (m)")
                _plt.ylabel(r"yp (rad)")
                startx, endx = ax4.get_xlim()
                starty, endy = ax4.get_ylim()
                ax4.xaxis.set_ticks([startx,0,endx])
                ax4.yaxis.set_ticks([starty,0,endy])
                locs,labels = _plt.xticks()
                _plt.xticks(locs, ["%2.1e" % x for x in locs])
                locs,labels = _plt.yticks()
                _plt.yticks(locs, ["%2.1e" % x for x in locs])
                ax4.tick_params(axis='both', which='major', pad=7)

                _plt.subplots_adjust(left=0.1,right=0.9,top=0.95, bottom=0.15, wspace=0.35, hspace=0.3)

                #PT vs T
                f = _plt.figure(self.figureNr+42, figsize=(13, 8), dpi=80, facecolor='w', edgecolor='k')
                f.suptitle(self.filename)
                _plt.clf()
                #X vs Y
                ax1 = _plt.gca()
                #arrow_width = abs(_np.min(fresy))*arrow_width_scale
                #_plt.quiver(Mx,My,-fresx,-fresy,angles='xy',scale_units='xy',scale=1,color='r',units='x',width=arrow_width,headwidth=3)
                #_plt.plot(Mx,My,'b.',label='PTC')
                #_plt.plot(Bx,By,'g.',label='BDSIM')
                _plt.scatter(MPT,MT,label='PTC',s=10,facecolors='red', edgecolors="red")
                _plt.scatter(BPT,Bt,label='BDSIM',s=60,facecolors='none',edgecolors="blue")
                if addPrimaries:
                    _plt.plot(Bx0,By0,'r.',label='BDSIM prim')
                _plt.legend(prop={'size':10})
                _plt.xlabel(r"(p-p0)/p0")
                _plt.ylabel(r"$\beta$cT (m)")
                startx, endx = ax1.get_xlim()
                starty, endy = ax1.get_ylim()
                ax1.xaxis.set_ticks([startx,0,endx])
                ax1.yaxis.set_ticks([starty,0,endy])
                locs,labels = _plt.xticks()
                _plt.xticks(locs, ["%2.1e" % x for x in locs])
                locs,labels = _plt.yticks()
                _plt.yticks(locs, ["%2.1e" % x for x in locs])
                ax1.tick_params(axis='both', which='major', pad=7)

            
                # 1d plots
                # x comparison
                f = _plt.figure(self.figureNr+6, figsize=(11, 8), dpi=80, facecolor='w', edgecolor='k')
                f.suptitle(self.filename)
                _plt.clf()

                nbinsx = _np.linspace(min(Mx),max(Mx),10)    #fix bins to avoid potential underflow/overflow
                nbinsy = _np.linspace(min(My),max(My),10)
                nbinsxp = _np.linspace(min(Mxp),max(Mxp),10)
                nbinsyp = _np.linspace(min(Myp),max(Myp),10)
                
                ax1 = f.add_subplot(221)
                ax1.hist(Mx,nbinsx,color='b',label='PTC',histtype='step')
                ax1.hist(Bx,nbinsx,color='g',label='BDSIM',histtype='step')
                if addPrimaries:
                    ax1.hist(Bx0,nbinsx,color='r',label='BDSIM prim',histtype='step')
                ax1.legend(fontsize='x-small',loc=0)
                ax1.set_xlabel(r"x (m)")
                startx, endx = ax1.get_xlim()
                starty, endy = ax1.get_ylim()
                ax1.xaxis.set_ticks([startx,0,endx])
                ax1.yaxis.set_ticks([starty,0,endy])
                
            
                # y comparison
                ax2 = f.add_subplot(222)
                ax2.hist(My,nbinsy,color='b',label='PTC',histtype='step')
                ax2.hist(By,nbinsy,color='g',label='BDSIM',histtype='step')
                if addPrimaries:
                    ax2.hist(By0,nbinsy,color='r',label='BDSIM prim',histtype='step')
                ax2.legend(fontsize='x-small',loc=0)
                ax2.set_xlabel(r"y (m)")
                startx, endx = ax2.get_xlim()
                starty, endy = ax2.get_ylim()
                ax2.xaxis.set_ticks([startx,0,endx])
                ax2.yaxis.set_ticks([starty,0,endy])

                # xp comparison
                ax3 = f.add_subplot(223)
                ax3.hist(Mxp,nbinsxp,color='b',label='PTC',histtype='step')
                ax3.hist(Bxp,nbinsxp,color='g',label='BDSIM',histtype='step')
                if addPrimaries:
                    ax3.hist(Bxp0,nbinsxp,color='r',label='BDSIM prim',histtype='step')
                ax3.legend(fontsize='x-small',loc=0)
                ax3.set_xlabel(r"x' (rad)")
                startx, endx = ax3.get_xlim()
                starty, endy = ax3.get_ylim()
                ax3.xaxis.set_ticks([startx,0,endx])
                ax3.yaxis.set_ticks([starty,0,endy])

                # yp comparison
                ax4 = f.add_subplot(224)
                ax4.hist(Myp,nbinsyp,color='b',label='PTC',histtype='step')
                ax4.hist(Byp,nbinsyp,color='g',label='BDSIM',histtype='step')
                if addPrimaries:
                    ax4.hist(Byp0,nbinsyp,color='r',label='BDSIM prim',histtype='step')
                ax4.legend(fontsize='x-small',loc=0)
                ax4.set_xlabel(r"y' (rad)")
                startx, endx = ax4.get_xlim()
                starty, endy = ax4.get_ylim()
                ax4.xaxis.set_ticks([startx,0,endx])
                ax4.yaxis.set_ticks([starty,0,endy])
            
                _plt.subplots_adjust(left=0.1,right=0.9,top=0.95, bottom=0.15, wspace=0.3, hspace=0.2)

            """
            # residuals in one plot
            nbins=50
            
            f = _plt.figure(self.figureNr+10, figsize=(13, 8), dpi=80, facecolor='w', edgecolor='k')
            _plt.clf()

            axX = f.add_subplot(221)
            hist, xedges, yedges = _np.histogram2d(Mx,fresx,bins=nbins)
            hist = _np.rot90(hist)                         #flip and rotate the plots to display them properly
            hist = _np.flipud(hist)
            histmasked = _np.ma.masked_where(hist==0,hist) # Mask pixels with a value of zero
            
            _plt.pcolormesh(xedges,yedges,histmasked)
            axX.set_xlabel('x(m)')
            axX.set_ylabel('$Residuals_{x}$(m)')
            cbar = _plt.colorbar()
            cbar.ax.set_ylabel('Counts')
            startx, endx = axX.get_xlim()
            starty, endy = axX.get_ylim()
            axX.xaxis.set_ticks([startx,0,endx])
            axX.yaxis.set_ticks([starty,0,endy])
            
            axX = f.add_subplot(222)
            hist, xedges, yedges = _np.histogram2d(My,fresy,bins=nbins)
            hist = _np.rot90(hist)
            hist = _np.flipud(hist)
            histmasked = _np.ma.masked_where(hist==0,hist)
            
            _plt.pcolormesh(xedges,yedges,histmasked)
            axX.set_xlabel('y(m)')
            axX.set_ylabel('$Residuals_{y}$(m)')
            cbar = _plt.colorbar()
            cbar.ax.set_ylabel('Counts')
            startx, endx = axX.get_xlim()
            starty, endy = axX.get_ylim()
            axX.xaxis.set_ticks([startx,0,endx])
            axX.yaxis.set_ticks([starty,0,endy])
            
            axX = f.add_subplot(223)
            hist, xedges, yedges = _np.histogram2d(Mxp,fresxp,bins=nbins)
            hist = _np.rot90(hist)
            hist = _np.flipud(hist)
            histmasked = _np.ma.masked_where(hist==0,hist)
            
            _plt.pcolormesh(xedges,yedges,histmasked)
            axX.set_xlabel('xp(rad)')
            axX.set_ylabel('$Residuals_{xp}$(rad)')
            cbar = _plt.colorbar()
            cbar.ax.set_ylabel('Counts')
            startx, endx = axX.get_xlim()
            starty, endy = axX.get_ylim()
            axX.xaxis.set_ticks([startx,0,endx])
            axX.yaxis.set_ticks([starty,0,endy])
            
            axX = f.add_subplot(224)
            hist, xedges, yedges = _np.histogram2d(Myp,fresyp,bins=nbins)
            hist = _np.rot90(hist)
            hist = _np.flipud(hist)
            histmasked = _np.ma.masked_where(hist==0,hist)
            
            _plt.pcolormesh(xedges,yedges,histmasked)
            axX.set_xlabel('yp(rad)')
            axX.set_ylabel('$Residuals_{yp}$(rad)')
            cbar = _plt.colorbar()
            cbar.ax.set_ylabel('Counts')
            startx, endx = axX.get_xlim()
            starty, endy = axX.get_ylim()
            axX.xaxis.set_ticks([startx,0,endx])
            axX.yaxis.set_ticks([starty,0,endy])

            _plt.subplots_adjust(left=0.15, right=0.95, top=0.95, bottom=0.15, wspace=0.39, hspace=0.3)
            """

            f = _plt.figure(self.figureNr+10, figsize=(13, 8), dpi=80, facecolor='w', edgecolor='k')
            _plt.clf()

            ax1 = f.add_subplot(221)

            binned    = _bin(Mx,fresx, 'mean', bins=15 )
            bin_means = binned[0]
            bin_edges = binned[1]
            bin_std   = _bin(Mx, fresx, _np.std, bins=15)[0]
            bin_count = _bin(Mx, fresx, 'count', bins=15)[0]
            bin_err   = _np.nan_to_num(bin_std/_np.sqrt(bin_count))
            
            ax1.bar(bin_edges[:-1], bin_means, 0.85*(bin_edges[1]-bin_edges[0]), yerr=bin_err, color='b',ec='b',alpha=0.5)
            ax1.set_xlim([min(Mx),max(Mx)])
            startx, endx = ax1.get_xlim()
            starty, endy = ax1.get_ylim()
            ax1.xaxis.set_ticks([startx,0,endx])
            ax1.yaxis.set_ticks([starty,0,endy])
            locs,labels = _plt.xticks()
            _plt.xticks(locs, ["%2.1e" % x for x in locs])
            locs,labels = _plt.yticks()
            _plt.yticks(locs, ["%2.1e" % x for x in locs])
            ax1.tick_params(axis='both', which='major', pad=7)
            ax1.set_xlabel('x(m)')
            ax1.set_ylabel('x residuals(m)')

            ax2 = f.add_subplot(222)

            binned    = _bin(My,fresy, 'mean', bins=15 )
            bin_means = binned[0]
            bin_edges = binned[1]
            bin_std   = _bin(My, fresy, _np.std, bins=15)[0]
            bin_count = _bin(My, fresy, 'count', bins=15)[0]
            bin_err   = _np.nan_to_num(bin_std/_np.sqrt(bin_count))
            
            ax2.bar(bin_edges[:-1], bin_means, 0.85*(bin_edges[1]-bin_edges[0]), yerr=bin_err, color='r',ec='r',alpha=0.5)
            ax2.set_xlim([min(My),max(My)])
            startx, endx = ax2.get_xlim()
            starty, endy = ax2.get_ylim()
            ax2.xaxis.set_ticks([startx,0,endx])
            ax2.yaxis.set_ticks([starty,0,endy])
            locs,labels = _plt.xticks()
            _plt.xticks(locs, ["%2.1e" % x for x in locs])
            locs,labels = _plt.yticks()
            _plt.yticks(locs, ["%2.1e" % x for x in locs])
            ax2.tick_params(axis='both', which='major', pad=7)
            ax2.set_xlabel('y(m)')
            ax2.set_ylabel('y residuals(m)')
            
            ax3 = f.add_subplot(223)

            binned    = _bin(Mxp,fresxp, 'mean', bins=15 )
            bin_means = binned[0]
            bin_edges = binned[1]
            bin_std   = _bin(Mxp, fresxp, _np.std, bins=15)[0]
            bin_count = _bin(Mxp, fresxp, 'count', bins=15)[0]
            bin_err   = _np.nan_to_num(bin_std/_np.sqrt(bin_count))
            
            ax3.bar(bin_edges[:-1], bin_means, 0.85*(bin_edges[1]-bin_edges[0]), yerr=bin_err, color='g',ec='g',alpha=0.5)
            ax3.set_xlim([min(Mxp),max(Mxp)])
            startx, endx = ax3.get_xlim()
            starty, endy = ax3.get_ylim()
            ax3.xaxis.set_ticks([startx,0,endx])
            ax3.yaxis.set_ticks([starty,0,endy])
            locs,labels = _plt.xticks()
            _plt.xticks(locs, ["%2.1e" % x for x in locs])
            locs,labels = _plt.yticks()
            _plt.yticks(locs, ["%2.1e" % x for x in locs])
            ax3.tick_params(axis='both', which='major', pad=7)
            ax3.set_xlabel('xp(rad)')
            ax3.set_ylabel('xp residuals(rad)')

            ax4 = f.add_subplot(224)

            binned    = _bin(Myp,fresyp, 'mean', bins=15 )
            bin_means = binned[0]
            bin_edges = binned[1]
            bin_std   = _bin(Myp, fresyp, _np.std, bins=15)[0]
            bin_count = _bin(Myp, fresyp, 'count', bins=15)[0]
            bin_err   = _np.nan_to_num(bin_std/_np.sqrt(bin_count))
            #print "total entries=",sum(bin_count),"bin counts= ", bin_count,"bin means= ", bin_means, "bin std= ", bin_err
            
            ax4.bar(bin_edges[:-1], bin_means, 0.85*(bin_edges[1]-bin_edges[0]), yerr=bin_err, color='c',ec='c',alpha=0.5)
            ax4.set_xlim([min(Myp),max(Myp)])
            startx, endx = ax4.get_xlim()
            starty, endy = ax4.get_ylim()
            ax4.xaxis.set_ticks([startx,0,endx])
            ax4.yaxis.set_ticks([starty,0,endy])
            locs,labels = _plt.xticks()
            _plt.xticks(locs, ["%2.1e" % x for x in locs])
            locs,labels = _plt.yticks()
            _plt.yticks(locs, ["%2.1e" % x for x in locs])
            ax4.tick_params(axis='both', which='major', pad=7)
            ax4.set_xlabel('yp(rad)')
            ax4.set_ylabel('yp residuals(rad)')

            _plt.subplots_adjust(left=0.15,right=0.95,top=0.95,wspace=0.39,hspace=0.25)
            
            if(showPlots):
                _plt.show()

        #Open pdf output file and save all plots to it
        pdf = _pdf.PdfPages(self.filename+"_plots.pdf")
        for i in _plt.get_fignums():
            pdf.savefig(i)
        pdf.close()
        



       

        

            

        
            


        

    




import numpy as _np
import pybdsim.Data as _Data

def BdsimElement2TransferMatrix(bdsfile,element,outputType="list"):
    """
    Calculates the Transfer Matrix of an element in a BDSIM beamline from an output (non-optical) 
    rootfile of the beamline by using a Singular Value Decomposition pseudo-inverse of a matrix of
    particle trajectories.
    
    Inputs:
    bdsfile - File to use to find the element.
    element - Number of element (must be >0, as requires an element before the element of interest).
    outputtype - "list", "dict" (default "list"). Controls output type, list returns a 16-element list of 
    floats of the matrix elements, dict returns a dict of labelled elements r11 through r44.
    """
    if element == 0:
        raise ValueError('Element to convert must have a preceding Element.')
    
    #take sampler data before and after element
    rootfile    = _Data.Load(bdsfile)
    priorund    = _Data.SamplerData(rootfile,element-1)
    postund     = _Data.SamplerData(rootfile,element)
    
    #put into matrix, transpose so each row is a single particle's co-ordinates.
    priormatrix = _np.asmatrix([priorund.data['x'], priorund.data['xp'], priorund.data['y'], priorund.data['yp']]).T
    lefthandx  = _np.asmatrix(postund.data['x']).T
    lefthandxp = _np.asmatrix(postund.data['xp']).T
    lefthandy  = _np.asmatrix(postund.data['y']).T
    lefthandyp = _np.asmatrix(postund.data['yp']).T
    
    #Pseudoinverse using SVD method
    righthandinverse = _np.linalg.pinv(priormatrix)
    
    #solve for matrix single row at a time for clarity
    xrow  = righthandinverse * lefthandx
    xprow = righthandinverse * lefthandxp
    yrow  = righthandinverse * lefthandy
    yprow = righthandinverse * lefthandyp
   
    if outputType == "dict" :
        output = {
            "r11" : xrow.item(0),
            "r12" : xrow.item(1),
            "r13" : xrow.item(2),
            "r14" : xrow.item(3),
            "r21" : xprow.item(0),
            "r22" : xprow.item(1),
            "r23" : xprow.item(2),
            "r24" : xprow.item(3),
            "r31" : yrow.item(0),
            "r32" : yrow.item(1),
            "r33" : yrow.item(2),
            "r34" : yrow.item(3),
            "r41" : yprow.item(0),
            "r42" : yprow.item(1),
            "r43" : yprow.item(2),
            "r44" : yprow.item(3),
                 }
    elif outputType == "list":
        output = [xrow.item(0),xrow.item(1),xrow.item(2),xrow.item(3),
                  xprow.item(0),xprow.item(1),xprow.item(2),xprow.item(3),
                  yrow.item(0),yrow.item(1),yrow.item(2),yrow.item(3),
                  yprow.item(0),yprow.item(1),yprow.item(2),yprow.item(3)]
    else:
        raise TypeError('outputType: "'+outputType+'" invalid.')
        
    return output
